# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class PipelineUtil:
    """Utility methods for pipeline analysis"""

    DEFAULT_EQ_FRAGILITY_KEY = "pgv"
    LIQ_FRAGILITY_KEY = "pgd"
    DEFAULT_TSU_FRAGILITY_KEY = "Non-Retrofit inundationDepth Fragility ID Code"

    @staticmethod
    def convert_result_unit(result_unit: str, result: float):
        """Convert values between different units.

        Args:
            result_unit (str): Resulting unit.
            result (float): Input value.

        Returns:
            float: Converted value.

        """

        if result_unit.lower() == "repairs/km":
            return result
        elif result_unit.lower() == "repairs/1000ft":
            return result / 0.3048

        print("Result type was not found so we didn't change it.  For pipes, all results should convert from their "
              "unit type into Repairs per Kilometer for uniform results.  We found a result type of " + result_unit)
        return result

    @staticmethod
    def get_pipe_length(pipeline):
        """Get pipe length.

        Args:
            pipeline (obj): A JSON-like description of pipeline properties.

        Returns:
            float: Pipe length.

        """
        pipe_length = 0.0

        if 'pipelength' in pipeline['properties']:
            pipe_length = float(pipeline['properties']['pipelength'])
        elif 'length_km' in pipeline['properties']:
            pipe_length = float(pipeline['properties']['length_km'])
        elif 'length' in pipeline['properties']:
            pipe_length = float(pipeline['properties']['length'])
        else:
            print("Pipeline has no length attribute")

        return pipe_length

    @staticmethod
    def get_pipe_diameter(pipeline):
        """Get pipe diameter.

        Args:
            pipeline (obj): A JSON-like description of pipeline properties.

        Returns:
            float: Pipe diameter.

        """
        diameter = 0.0
        if 'diameter' in pipeline['properties']:
            diameter = float(pipeline['properties']['diameter'])

        return diameter



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import os

import pyomo.environ as pyo
from pyomo.core import value

from pyincore.analyses.indp.infrastructureutil import InfrastructureUtil
from pyincore.analyses.indp.indpresults import INDPResults
import pandas as pd

class INDPUtil:

    @staticmethod
    def get_resource_suffix(params):
        """
        This function generates the part of the suffix of result folders that pertains to resource cap(s).

        Args:
            params (dict): Parameters that are needed to run the INDP optimization.

        Returns:
            out_dir_suffix_res (str): The part of the suffix of result folders that pertains to resource cap(s).

        """
        out_dir_suffix_res = ''
        for rc, val in params["V"].items():
            if isinstance(val, int):
                if rc != '':
                    out_dir_suffix_res += rc[0] + str(val)
                else:
                    out_dir_suffix_res += str(val)
            else:
                out_dir_suffix_res += rc[0] + str(sum([lval for _, lval in val.items()])) + '_fixed_layer_Cap'
        return out_dir_suffix_res

    @staticmethod
    def time_resource_usage_curves(power_arcs, power_nodes, water_arcs, water_nodes, wf_restoration_time_sample,
                                   wf_repair_cost_sample, pipeline_restoration_time, pipeline_repair_cost,
                                   epf_restoration_time_sample, epf_repair_cost_sample):
        """
        This module calculates the repair time for nodes and arcs for the current scenario based on their damage
        state, and writes them to the input files of INDP. Currently, it is only compatible with NIST testbeds.

        Args:
            power_arcs (dataframe):
            power_nodes (dataframe):
            water_arcs (dataframe):
            water_nodes (dataframe):
            wf_restoration_time_sample (dataframe):
            wf_repair_cost_sample (dataframe):
            pipeline_restoration_time (dataframe):
            pipeline_repair_cost (dataframe):
            epf_restoration_time_sample (dataframe):
            epf_repair_cost_sample (dataframe):

        Returns:
            water_nodes:
            water_arcs:
            power_nodes:
            power_arcs:

        """
        _water_nodes = water_nodes.merge(wf_repair_cost_sample, on='guid', how='left').merge(
            wf_restoration_time_sample, on='guid', how='left')
        water_nodes['p_time'] = _water_nodes['repairtime']
        water_nodes['p_time'].fillna(0, inplace=True)
        water_nodes['p_budget'] = _water_nodes['budget']
        water_nodes['p_budget'].fillna(0, inplace=True)
        water_nodes['q'] = _water_nodes['repaircost']
        water_nodes['q'].fillna(0, inplace=True)

        _power_nodes = power_nodes.merge(epf_repair_cost_sample, on='guid', how='left').merge(
            epf_restoration_time_sample, on='guid', how='left')
        power_nodes['p_time'] = _power_nodes['repairtime']
        power_nodes['p_time'].fillna(0, inplace=True)
        power_nodes['p_budget'] = _power_nodes['budget']
        power_nodes['p_budget'].fillna(0, inplace=True)
        power_nodes['q'] = _power_nodes['repaircost']
        power_nodes['q'].fillna(0, inplace=True)

        _water_arcs = water_arcs.merge(pipeline_repair_cost, on='guid', how='left').merge(
            pipeline_restoration_time, on='guid', how='left')
        water_arcs['h_time'] = _water_arcs["repair_time"]
        water_arcs['h_time'].fillna(0, inplace=True)
        water_arcs['h_budget'] = _water_arcs['budget'].astype(float)
        water_arcs['h_budget'].fillna(0, inplace=True)
        water_arcs['f'] = _water_arcs['repaircost'].astype(float)
        water_arcs['f'].fillna(0, inplace=True)

        power_arcs['h_time'] = 0
        power_arcs['h_budget'] = 0
        power_arcs['f'] = 0

        return water_nodes, water_arcs, power_nodes, power_arcs

    @staticmethod
    def initialize_network(power_nodes, power_arcs, water_nodes, water_arcs, interdep, cost_scale=1.0,
                           extra_commodity=None):
        """
        This function initializes a :class:`~infrastructure.InfrastructureNetwork` object based on network data.

        Args:
            cost_scale (float): Scales the cost to improve efficiency. The default is 1.0:
            extra_commodity (dict): Dictionary of commodities other than the default one for each layer of the
            network. The default is 'None', which means that there is only one commodity per layer.

        Returns:
            interdep_net (class):`~infrastructure.InfrastructureNetwork` The object containing the network data.

        """
        interdep_net = InfrastructureUtil.load_infrastructure_array_format_extended(power_nodes,
                                                                                    power_arcs,
                                                                                    water_nodes,
                                                                                    water_arcs,
                                                                                    interdep,
                                                                                    cost_scale=cost_scale,
                                                                                    extra_commodity=extra_commodity)
        return interdep_net

    @staticmethod
    def save_indp_model_to_file(model, out_model_dir, t, layer=0, suffix=''):
        """
        This function saves pyomo optimization model to file.

        Parameters
        ----------
        model : Pyomo.Model
            Pyomo optimization model
        out_model_dir : str
            Directory to which the models should be written
        t : int
            The time step corresponding to the model
        layer : int
            The layer number corresponding to the model. The default is 0, which means the model includes all layers in
            the analysis
        suffix : str
            The suffix that should be added to files when saved. The default is ''.
        Returns
        -------
        None.

        """
        if not os.path.exists(out_model_dir):
            os.makedirs(out_model_dir)
            # Write models to file
        l_name = "/Model_t%d_l%d_%s.txt" % (t, layer, suffix)
        file_id = open(out_model_dir + l_name, 'w')
        model.pprint(ostream=file_id)
        file_id.close()
        # Write solution to file
        s_name = "/Solution_t%d_l%d_%s.txt" % (t, layer, suffix)
        file_id = open(out_model_dir + s_name, 'w')
        for vv in model.component_data_objects(pyo.Var):
            if vv.value:
                file_id.write('%s %g\n' % (str(vv), vv.value))
            else:
                file_id.write('%s NONE\n' % (str(vv)))
        file_id.write('Obj: %g' % value(model.Obj))
        file_id.close()

    @staticmethod
    def apply_recovery(N, indp_results, t):
        """
        This function applies the restoration decisions (solution of INDP) to a Gurobi model by changing the state of
        repaired elements to functional

        Parameters
        ----------
        N : :class:`~infrastructure.InfrastructureNetwork`
            The model of the interdependent network.
        indp_results : INDPResults
            A :class:`~indputils.INDPResults` object containing the optimal restoration decisions.
        t : int
            The time step to which the results should apply.

        Returns
        -------
        None.

        """
        for action in indp_results[t]['actions']:
            if "/" in action:
                # Edge recovery action.
                data = action.split("/")
                src = tuple([int(x) for x in data[0].split(".")])
                dst = tuple([int(x) for x in data[1].split(".")])
                N.G[src][dst]['data']['inf_data'].functionality = 1.0
            else:
                # Node recovery action.
                node = tuple([int(x) for x in action.split(".")])
                # print "Applying recovery:",node
                N.G.nodes[node]['data']['inf_data'].repaired = 1.0
                N.G.nodes[node]['data']['inf_data'].functionality = 1.0

    @staticmethod
    def collect_results(model, controlled_layers, coloc=True):
        """
        This function computes the results (actions and costs) of the optimal results and writes them to a
        :class:`~indputils.INDPResults` object.

        Parameters
        ----------
        model : pyomo.model
            The object containing the the solved optimization problem.
        controlled_layers : list
            Layer IDs that can be recovered in this optimization.
        coloc : bool, optional
            If false, exclude geographical interdependency from the results. The default is True.

        Returns
        -------
        indp_results : INDPResults
        A :class:`~indputils.INDPResults` object containing the optimal restoration decisions.

        """
        layers = controlled_layers
        indp_results = INDPResults(layers)
        # compute total demand of all layers and each layer
        total_demand = 0.0
        total_demand_layer = {layer: 0.0 for layer in layers}
        for n, d in model.n_hat.nodes(data=True):
            demand_value = d['data']['inf_data'].demand
            if demand_value < 0:
                total_demand += demand_value
                total_demand_layer[n[1]] += demand_value
        for t in range(model.T):
            node_cost = 0.0
            arc_cost = 0.0
            flow_cost = 0.0
            over_supp_cost = 0.0
            under_supp_cost = 0.0
            under_supp = 0.0
            space_prep_cost = 0.0
            node_cost_layer = {layer: 0.0 for layer in layers}
            arc_cost_layer = {layer: 0.0 for layer in layers}
            flow_cost_layer = {layer: 0.0 for layer in layers}
            over_supp_cost_layer = {layer: 0.0 for layer in layers}
            under_supp_cost_layer = {layer: 0.0 for layer in layers}
            under_supp_layer = {layer: 0.0 for layer in layers}
            space_prep_cost_layer = {layer: 0.0 for layer in layers}  # !!! populate this for each layer
            # Record node recovery actions.
            for n in model.n_hat_prime_nodes:
                if model.T == 1:
                    node_state = model.w[n, t].value
                else:
                    node_state = model.w_tilde[n, t].value
                if node_state == 1:
                    action = str(n[0]) + "." + str(n[1])
                    indp_results.add_action(t, action)
            # Record edge recovery actions.
            for i, k, j, kb in model.a_hat_prime:
                if model.T == 1:
                    arc_state = model.y[i, k, j, kb, t].value
                else:
                    arc_state = model.y_tilde[i, k, j, kb, t].value
                if arc_state == 1:
                    action = str(i) + "." + str(k) + "/" + str(j) + "." + str(kb)
                    indp_results.add_action(t, action)
            # Calculate space preparation costs.
            if coloc:
                for s in model.S.value:
                    if model.z[s.id, t].value:
                        space_prep_cost += s.cost * model.z[s.id, t].value
            indp_results.add_cost(t, "Space Prep", space_prep_cost, space_prep_cost_layer)
            # Calculate arc preparation costs.
            for i, k, j, kb in model.a_hat_prime:
                a = model.n_hat[i, k][j, kb]['data']['inf_data']
                if model.T == 1:
                    arc_state = model.y[i, k, j, kb, t].value
                else:
                    arc_state = model.y_tilde[i, k, j, kb, t].value
                arc_cost += (a.reconstruction_cost / 2.0) * arc_state
                arc_cost_layer[k] += (a.reconstruction_cost / 2.0) * arc_state
            indp_results.add_cost(t, "Arc", arc_cost, arc_cost_layer)
            # Calculate node preparation costs.
            for n in model.n_hat_prime_nodes:
                d = model.n_hat.nodes[n]['data']['inf_data']
                if model.T == 1:
                    node_state = model.w[n, t].value
                else:
                    node_state = model.w_tilde[n, t].value
                node_cost += d.reconstruction_cost * node_state
                node_cost_layer[n[1]] += d.reconstruction_cost * node_state
            indp_results.add_cost(t, "Node", node_cost, node_cost_layer)
            # Calculate under/oversupply costs.
            for n, d in model.n_hat.nodes(data=True):
                over_supp_cost += d['data']['inf_data'].oversupply_penalty * model.delta_p[n, 'b', t].value
                over_supp_cost_layer[n[1]] += d['data']['inf_data'].oversupply_penalty * model.delta_p[n, 'b', t].value
                under_supp += model.delta_m[n, 'b', t].value
                under_supp_layer[n[1]] += model.delta_m[n, 'b', t].value / total_demand_layer[n[1]]
                under_supp_cost += d['data']['inf_data'].undersupply_penalty * model.delta_m[n, 'b', t].value
                under_supp_cost_layer[n[1]] += d['data']['inf_data'].undersupply_penalty * model.delta_m[
                    n, 'b', t].value
            indp_results.add_cost(t, "Over Supply", over_supp_cost, over_supp_cost_layer)
            indp_results.add_cost(t, "Under Supply", under_supp_cost, under_supp_cost_layer)
            indp_results.add_cost(t, "Under Supply Perc", under_supp / total_demand, under_supp_layer)
            # Calculate flow costs.
            for i, k, j, kb in model.a_hat:
                a = model.n_hat[i, k][j, kb]['data']['inf_data']
                flow_cost += a.flow_cost * model.x[i, k, j, kb, 'b', t].value
                flow_cost_layer[k] += a.flow_cost * model.x[i, k, j, kb, 'b', t].value
            indp_results.add_cost(t, "Flow", flow_cost, flow_cost_layer)
            # Calculate total costs.
            total_lyr = {}
            total_nd_lyr = {}
            for layer in layers:
                total_lyr[layer] = flow_cost_layer[layer] + arc_cost_layer[layer] + node_cost_layer[layer] + \
                                   over_supp_cost_layer[layer] + under_supp_cost_layer[layer] + \
                                   space_prep_cost_layer[layer]
                total_nd_lyr[layer] = space_prep_cost_layer[layer] + arc_cost_layer[layer] + flow_cost \
                                      + node_cost_layer[layer]
            indp_results.add_cost(t, "Total", flow_cost + arc_cost + node_cost + over_supp_cost + under_supp_cost +
                                  space_prep_cost, total_lyr)
            indp_results.add_cost(t, "Total no disconnection", space_prep_cost + arc_cost + flow_cost + node_cost,
                                  total_nd_lyr)
        return indp_results

    @staticmethod
    def initial_state_node_rule(model, i, k):
        return model.w[i, k, 0] == 0.0

    @staticmethod
    def initial_state_arc_rule(model, i, k, j, kb):
        return model.y[i, k, j, kb, 0] == 0.0

    @staticmethod
    def time_dependent_node_rule(model, i, k, t):
        if t > 0:
            return sum(model.w_tilde[i, k, t_p] for t_p in range(1, t + 1) if t > 0) >= model.w[i, k, t]
        else:
            return pyo.Constraint.Skip

    @staticmethod
    def time_dependent_arc_rule(model, i, k, j, kb, t):
        if t > 0:
            return sum(model.y_tilde[i, k, j, kb, t_p] for t_p in range(1, t + 1) if t > 0) >= model.y[i, k, j, kb, t]
        else:
            return pyo.Constraint.Skip

    @staticmethod
    def arc_equality_rule(model, i, k, j, kb, t):
        try:
            if model.T == 1:
                return model.y[i, k, j, kb, t] == model.y[j, kb, i, k, t]
            else:
                return model.y[i, k, j, kb, t] == model.y[j, kb, i, k, t], \
                       model.y_tilde[i, k, j, kb, t] == model.y_tilde[j, kb, i, k, t]
        except KeyError:
            return pyo.Constraint.Skip

    @staticmethod
    def flow_conserv_node_rule(model, i, k, layer, t):
        d = model.n_hat.nodes[(i, k)]
        out_flow_constr = 0
        in_flow_constr = 0
        demand_constr = 0
        for u, v, a in model.n_hat.out_edges((i, k), data=True):
            if layer == 'b' or layer in a['data']['inf_data'].extra_com.keys():
                out_flow_constr += model.x[u, v, layer, t]
        for u, v, a in model.n_hat.in_edges((i, k), data=True):
            if layer == 'b' or layer in a['data']['inf_data'].extra_com.keys():
                in_flow_constr += model.x[u, v, layer, t]
        if layer == 'b':
            demand_constr += d['data']['inf_data'].demand - model.delta_p[i, k, layer, t] \
                             + model.delta_m[i, k, layer, t]
        else:
            demand_constr += d['data']['inf_data'].extra_com[layer]['demand'] - model.delta_p[i, k, layer, t] + \
                             model.delta_m[i, k, layer, t]
        return out_flow_constr - in_flow_constr == demand_constr

    @staticmethod
    def flow_in_functionality_rule(model, i, k, j, kb, t):
        if not model.functionality:
            interdep_nodes_list = model.interdep_nodes.keys()  # Interdependent nodes with a damaged dependee node
        else:
            interdep_nodes_list = model.interdep_nodes[t].keys()  # Interdependent nodes with a damaged dependee node
        a = model.n_hat[i, k][j, kb]['data']['inf_data']
        lhs = model.x[i, k, j, kb, 'b', t]
        for layer in a.extra_com.keys():
            lhs += model.x[i, k, j, kb, layer, t]
        if ((i, k) in model.n_hat_prime_nodes) | ((i, k) in interdep_nodes_list):
            return lhs <= a.capacity * model.w[i, k, t]
        else:
            return lhs <= a.capacity * model.N.G.nodes[(i, k)]['data']['inf_data'].functionality

    @staticmethod
    def flow_out_functionality_rule(model, i, k, j, kb, t):
        if not model.functionality:
            interdep_nodes_list = model.interdep_nodes.keys()  # Interdependent nodes with a damaged dependee node
        else:
            interdep_nodes_list = model.interdep_nodes[t].keys()  # Interdependent nodes with a damaged dependee node
        a = model.n_hat[i, k][j, kb]['data']['inf_data']
        lhs = model.x[i, k, j, kb, 'b', t]
        for layer in a.extra_com.keys():
            lhs += model.x[i, k, j, kb, layer, t]
        if ((j, kb) in model.n_hat_prime_nodes) | ((j, kb) in interdep_nodes_list):
            return lhs <= a.capacity * model.w[j, kb, t]
        else:
            return lhs <= a.capacity * model.N.G.nodes[(j, kb)]['data']['inf_data'].functionality

    @staticmethod
    def flow_arc_functionality_rule(model, i, k, j, kb, t):
        if not model.functionality:
            interdep_nodes_list = model.interdep_nodes.keys()  # Interdependent nodes with a damaged dependee node
        else:
            interdep_nodes_list = model.interdep_nodes[t].keys()  # Interdependent nodes with a damaged dependee node
        a = model.n_hat[i, k][j, kb]['data']['inf_data']
        lhs = model.x[i, k, j, kb, 'b', t]
        for layer in a.extra_com.keys():
            lhs += model.x[i, k, j, kb, layer, t]
        if (i, k, j, kb) in model.a_hat_prime:
            return lhs <= a.capacity * model.y[i, k, j, kb, t]
        else:
            return lhs <= a.capacity * model.N.G[(i, k)][(j, kb)]['data']['inf_data'].functionality

    @staticmethod
    def resource_rule(model, rc, t):
        resource_dict = model.v_r[rc]
        is_sep_res = False
        if isinstance(resource_dict, int):
            total_resource = resource_dict
        else:
            is_sep_res = True
            total_resource = sum([lval for _, lval in resource_dict.items()])
            assert len(resource_dict.keys()) == len(model.layers), "The number of resource  values does not match the \
                                                                    number of layers."
        resource_left_constr = 0
        if is_sep_res:
            res_left_constr_sep = {key: 0 for key in resource_dict.keys()}
        for i, k, j, kb in model.a_hat_prime:
            a = model.n_hat[i, k][j, kb]['data']['inf_data']
            idx_lyr = a.layer
            res_use = 0.5 * a.resource_usage['h_' + rc]
            if model.T == 1:
                resource_left_constr += res_use * model.y[i, k, j, kb, t]
                if is_sep_res:
                    res_left_constr_sep[idx_lyr] += res_use * model.y[i, k, j, kb, t]
            else:
                resource_left_constr += res_use * model.y_tilde[i, k, j, kb, t]
                if is_sep_res:
                    res_left_constr_sep[idx_lyr] += res_use * model.y_tilde[i, k, j, kb, t]
        for n in model.n_hat_prime_nodes:
            idx_lyr = n[1]
            d = model.n_hat.nodes[n]['data']['inf_data']
            res_use = d.resource_usage['p_' + rc]
            if model.T == 1:
                resource_left_constr += res_use * model.w[n, t]
                if is_sep_res:
                    res_left_constr_sep[idx_lyr] += res_use * model.w[n, t]
            else:
                resource_left_constr += res_use * model.w_tilde[n, t]
                if is_sep_res:
                    res_left_constr_sep[idx_lyr] += res_use * model.w_tilde[n, t]
        if not isinstance(resource_left_constr, int):
            if not is_sep_res:
                return resource_left_constr <= total_resource
            else:
                return resource_left_constr <= total_resource, [res_left_constr_sep[key] <= lval for key, lval in
                                                                resource_dict]
        else:
            return pyo.Constraint.Skip

    @staticmethod
    def interdependency_rule(model, i, k, t):
        if not model.functionality:
            if (i, k) in model.interdep_nodes:
                interdep_l_constr = 0
                interdep_r_constr = 0
                for interdep in model.interdep_nodes[(i, k)]:
                    src = interdep[0]
                    gamma = interdep[1]
                    if not model.n_hat.has_node(src):
                        interdep_l_constr += 0
                    else:
                        interdep_l_constr += model.w[src, t] * gamma
                interdep_r_constr += model.w[i, k, t]
                return interdep_l_constr >= interdep_r_constr
        else:
            if (i, k) in model.interdep_nodes[t]:
                # print interdep_nodes[t]
                interdep_l_constr = 0
                interdep_r_constr = 0
                for interdep in model.interdep_nodes[t][(i, k)]:
                    src = interdep[0]
                    gamma = interdep[1]
                    if not model.n_hat.has_node(src):
                        print("Forcing", str((i, k)), "to be 0 (dep. on", str(src), ")")
                        interdep_l_constr += 0
                    else:
                        interdep_l_constr += model.w[src, t] * gamma
                interdep_r_constr += model.w[i, k, t]
                return interdep_l_constr >= interdep_r_constr
        return pyo.Constraint.Skip

    @staticmethod
    def node_geographic_space_rule(model, s, i, k, t):
        d = model.n_hat.nodes[(i, k)]['data']['inf_data']
        if d.in_space(s):
            if model.T == 1:
                return model.w[(i, k), t] * d.in_space(s) <= model.z[s, t]
            else:
                return model.w_tilde[(i, k), t] * d.in_space(s) <= model.z[s, t]
        return pyo.Constraint.Skip

    @staticmethod
    def arc_geographic_space_rule(model, s, i, k, j, kb, t):
        a = model.n_hat[i, k][j, kb]['data']['inf_data']
        if a.in_space(s):
            if model.T == 1:
                return model.y[i, k, j, kb, t] * a.in_space(s) <= model.z[s, t]
            else:
                return model.y_tilde[i, k, j, kb, t] * a.in_space(s) <= model.z[s, t]
        return pyo.Constraint.Skip

    @staticmethod
    def collect_solution_pool(m, T, n_hat_prime, a_hat_prime):
        """
        This function collects the result (list of repaired nodes and arcs) for all feasible solutions in the
        solution pool.

        Parameters
        ----------
        m : gurobi.Model
            The object containing the solved optimization problem.
        T : int
            Number of time steps in the optimization (T=1 for iINDP, and T>=1 for td-INDP).
        n_hat_prime : list
            List of damaged nodes in controlled networks.
        a_hat_prime : list
            List of damaged arcs in controlled networks.

        Returns
        -------
        sol_pool_results : dict
        A dictionary containing one dictionary per solution that contains list of repaired node and arcs in the
        solution.

        """
        sol_pool_results = {}
        current_sol_count = 0
        for sol in range(m.SolCount):
            m.setParam('SolutionNumber', sol)
            # print(m.PoolObjVal)
            sol_pool_results[sol] = {'nodes': [], 'arcs': []}
            for t in range(T):
                # Record node recovery actions.
                for n, d in n_hat_prime:
                    node_var = 'w_tilde_' + str(n) + "," + str(t)
                    if T == 1:
                        node_var = 'w_' + str(n) + "," + str(t)
                    if round(m.getVarByName(node_var).xn) == 1:
                        sol_pool_results[sol]['nodes'].append(n)
                # Record edge recovery actions.
                for u, v, a in a_hat_prime:
                    arc_var = 'y_tilde_' + str(u) + "," + str(v) + "," + str(t)
                    if T == 1:
                        arc_var = 'y_' + str(u) + "," + str(v) + "," + str(t)
                    if round(m.getVarByName(arc_var).x) == 1:
                        sol_pool_results[sol]['arcs'].append((u, v))
            if sol > 0 and sol_pool_results[sol] == sol_pool_results[current_sol_count]:
                del sol_pool_results[sol]
            elif sol > 0:
                current_sol_count = sol
        return sol_pool_results

    @staticmethod
    def generate_intial_node_failure_state(wf_failure_state_df, epf_failure_state_df, water_nodes,
                                           power_nodes, sample_range):
        # todo change those hard coding
        network_name = {'Water': 1, 'Power': 3}
        combined_node_failed_states = pd.DataFrame()
        for node_fail_state, node_data, network_code in zip([wf_failure_state_df, epf_failure_state_df],
                                                            [water_nodes, power_nodes],
                                                            network_name.values()):
            node_fail_state[[str(x) for x in sample_range]] = \
                node_fail_state['failure'].str.split(',', expand=True).iloc[:, sample_range.start:sample_range.stop]
            node_fail_state = node_fail_state.drop(columns=['failure'])
            node_fail_state['name'] = 'nan'
            for index, row in node_data.iterrows():
                node_name = '(' + str(int(row['nodenwid'])) + ',' + str(network_code) + ')'
                if not pd.isna(row['guid']):
                    node_fail_state.loc[node_fail_state['guid'] == row['guid'], 'name'] = node_name
                else:
                    temp_dict = {**{'name': node_name, 'guid': 'nan'}, **{str(x): 1 for x in sample_range}}
                    node_fail_state = pd.concat([node_fail_state, pd.DataFrame([temp_dict])], ignore_index=True)
            combined_node_failed_states = pd.concat([combined_node_failed_states, node_fail_state])

        return combined_node_failed_states.dropna(subset=["name"])

    @staticmethod
    def generate_intial_link_failure_state(pipeline_failure_state_df, water_arcs, power_arcs, sample_range):
        # TODO remove hardcoded
        network_name = {'Water': 1, 'Power': 3}
        combined_arc_failed_states = pd.DataFrame()

        powerline_failure_state_df = pd.DataFrame()
        powerline_failure_state_df['name'] = 'nan'
        for index, row in power_arcs.iterrows():
            temp_dict = {str(x): 1 for x in sample_range}
            temp_dict['guid'] = row['guid']
            powerline_failure_state_df = pd.concat([powerline_failure_state_df, pd.DataFrame([temp_dict])],
                                                   ignore_index=True)
            s_node = int(row['fromnode'])
            e_node = int(row['tonode'])
            arc_name = '((' + str(s_node) + ',' + str(network_name["Power"]) + '),(' + str(e_node) + ',' + str(
                network_name["Power"]) + '))'
            powerline_failure_state_df.loc[powerline_failure_state_df['guid'] == row['guid'], 'name'] = arc_name

        pipeline_failure_state_df[[str(x) for x in sample_range]] = \
            pipeline_failure_state_df['failure'].str.split(',', expand=True).iloc[:,
            sample_range.start:sample_range.stop]
        pipeline_failure_state_df['name'] = 'nan'
        pipeline_failure_state_df = pipeline_failure_state_df.drop(columns=['failure'])
        for index, row in water_arcs.iterrows():
            s_node = int(row['fromnode'])
            e_node = int(row['tonode'])
            arc_name = '((' + str(s_node) + ',' + str(network_name["Water"]) + '),(' + str(e_node) + ',' + str(
                network_name["Water"]) + '))'
            if not pd.isna(row['guid']):
                pipeline_failure_state_df.loc[pipeline_failure_state_df['guid'] == row['guid'], 'name'] = arc_name
            else:
                temp_dict = {**{'name': arc_name, 'guid': 'nan'}, **{str(x): 1 for x in sample_range}}
                pipeline_failure_state_df = pd.concat([pipeline_failure_state_df, pd.DataFrame([temp_dict])],
                                                      ignore_index=True)

        combined_arc_failed_states = pd.concat([combined_arc_failed_states, powerline_failure_state_df])
        combined_arc_failed_states = pd.concat([combined_arc_failed_states, pipeline_failure_state_df])

        return combined_arc_failed_states.dropna(subset=["name"])

    @staticmethod
    def generate_distribution_nodes(water_arc_df, water_nodes_df):
        dist_nodes = list(set().union(water_arc_df['tonode'].unique(), water_arc_df['fromnode'].unique()))
        for node in dist_nodes:
            if node not in water_nodes_df["nodenwid"]:
                water_nodes_df = pd.concat([water_nodes_df, pd.DataFrame([{'utilfcltyc': 'Distribution Node',
                                                                          'nodenwid': node}])], ignore_index=True)
        return water_nodes_df



# Copyright (c) 2022 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import copy

import numpy as np
import pandas as pd
import networkx as nx
from typing import List
from pyincore import BaseAnalysis, NetworkDataset
from pyincore.analyses.epnfunctionality.epnfunctionalityutil import EpnFunctionalityUtil


class EpnFunctionality(BaseAnalysis):
    """Computes electric power infrastructure functionality.
    Args:
        incore_client: Service client with authentication info
    """

    def __init__(self, incore_client):
        super(EpnFunctionality, self).__init__(incore_client)

    def run(self):
        """Execute electric power facility functionality analysis """

        # get network dataset
        network_dataset = NetworkDataset.from_dataset(self.get_input_dataset("epn_network"))
        links_epl_gdf = network_dataset.links.get_dataframe_from_shapefile()
        nodes_epf_gdf = network_dataset.nodes.get_dataframe_from_shapefile()
        links_epl_gdf['weight'] = links_epl_gdf.loc[:, 'length_km']
        G_ep = network_dataset.get_graph_networkx()

        # get epf sample
        epf_dmg_fs = self.get_input_dataset('epf_sample_failure_state').get_dataframe_from_csv()
        epf_sample_df = pd.DataFrame(
            np.array([np.array(epf_dmg_fs.failure.values[i].split(',')).astype('int')
                      for i in np.arange(epf_dmg_fs.shape[0])]),
            index=epf_dmg_fs.guid.values)
        # get the sample number
        num_samples = epf_sample_df.shape[1]
        sampcols = ['s' + samp for samp in np.arange(num_samples).astype(str)]

        # add column
        epf_sample_df.columns = sampcols
        epf_sample_df1 = nodes_epf_gdf.loc[:, ['guid', 'nodenwid']].set_index('guid').join(epf_sample_df)

        # get gate station nodes
        gate_station_node_list = self.get_parameter("gate_station_node_list")
        if gate_station_node_list is None:
            # default to EPPL
            gatestation_nodes_class = 'EPPL'
            # get the guid from the matching class
            gate_station_node_list = nodes_epf_gdf[nodes_epf_gdf["utilfcltyc"] == gatestation_nodes_class]["nodenwid"]\
                .to_list()

        # calculate the distribution nodes
        distribution_sub_nodes = list(set(list(G_ep.nodes)) - set(gate_station_node_list))

        (fs_results, fp_results) = self.epf_functionality(distribution_sub_nodes, gate_station_node_list, num_samples,
                                                          sampcols, epf_sample_df1, G_ep)

        self.set_result_csv_data("sample_failure_state",
                                 fs_results, name=self.get_parameter("result_name") + "_failure_state",
                                 source="dataframe")
        self.set_result_csv_data("failure_probability",
                                 fp_results,
                                 name=self.get_parameter("result_name") + "_failure_probability",
                                 source="dataframe")

        return True

    def epf_functionality(self, distribution_sub_nodes, gate_station_node_list, num_samples, sampcols, epf_sample_df1,
                          G_ep):
        """
        Run EPN functionality analysis.

        Args:
            distribution_sub_nodes (list): distribution nodes
            gate_station_node_list (list): gate station nodes
            num_samples (int): number of simulations
            sampcols (list): list of number samples. e.g. "s0, s1,..."
            epf_sample_df1 (dataframe): epf mcs failure sample dataframe with added field "weight"
            G_ep (networkx object): constructed network

        Returns:
            fs_results (list): A list of dictionary with id/guid and failure state for N samples
            fp_results (list): A list dictionary with failure probability and other data/metadata.

        """

        # a distance of M denotes disconnection
        M = 9999
        func_ep_df = pd.DataFrame(np.zeros((len(distribution_sub_nodes), num_samples)), index=distribution_sub_nodes,
                                  columns=sampcols)

        for si, scol in enumerate(sampcols):
            nodestate_ep = epf_sample_df1.loc[:, ['nodenwid', scol]]
            linkstate_ep = None
            badlinks_ep = EpnFunctionalityUtil.get_bad_edges(G_ep, nodestate_ep, linkstate_ep, scol)
            badlinkdict_ep = {k: {'weight': M} for k in badlinks_ep}
            G1_ep = copy.deepcopy(G_ep)
            nx.set_edge_attributes(G1_ep, badlinkdict_ep)
            res_ep = EpnFunctionalityUtil.network_shortest_paths(G1_ep, gate_station_node_list, distribution_sub_nodes)
            func_ep_df.loc[distribution_sub_nodes, scol] = (res_ep < M) * 1

        # use nodenwid index to get its guid
        fs_temp = pd.merge(func_ep_df, epf_sample_df1["nodenwid"], left_index=True, right_on="nodenwid",
                           how='left').drop(columns=["nodenwid"])
        fp_temp = fs_temp.copy(deep=True)

        # shape the dataframe into failure probability and failure samples
        fs_temp['failure'] = fs_temp.astype(str).apply(','.join, axis=1)
        fs_results = fs_temp.filter(['failure'])
        fs_results.reset_index(inplace=True)
        fs_results = fs_results.rename(columns={'index': 'guid'})

        # calculate failure probability
        # count of 0s divided by sample size
        fp_temp["failure_probability"] = (num_samples - fp_temp.sum(axis=1).astype(int)) / num_samples
        fp_results = fp_temp.filter(['failure_probability'])
        fp_results.reset_index(inplace=True)
        fp_results = fp_results.rename(columns={'index': 'guid'})

        return fs_results, fp_results

    def get_spec(self):
        """Get specifications of the EPN functionality analysis.
        Returns:
            obj: A JSON object of specifications of the EPN functionality analysis.
        """
        return {
            'name': 'epn-functionality',
            'description': 'electric power network functionality analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'gate_station_node_list',
                    'required': False,
                    'description': "list of gate station nodes",
                    'type': List[int]
                },
            ],
            'input_datasets': [
                {
                    'id': 'epn_network',
                    'required': True,
                    'description': 'EPN Network Dataset',
                    'type': ['incore:epnNetwork'],
                },
                {
                    'id': 'epf_sample_failure_state',
                    'required': True,
                    'description': 'CSV file of failure state for each sample. Output from MCS analysis',
                    'type': ['incore:sampleFailureState']
                },
            ],
            'output_datasets': [
                {
                    'id': 'failure_probability',
                    'description': 'CSV file of failure probability',
                    'type': 'incore:failureProbability'
                },
                {
                    'id': 'sample_failure_state',
                    'description': 'CSV file of failure state for each sample',
                    'type': 'incore:sampleFailureState'
                },
            ]
        }



"""
functions for output validation

# NOTE:
This is a hack to create dataframes that resemble matrix results
returned in GAMS
"""

import pandas as pd


# Return initial values function
def base_value(vars, eqName):
    if vars.get_info(eqName)['size'] == 1:
        basevalues = vars.initialVals[vars.get_index(eqName)]

    else:
        rows = vars.get_info(eqName)['rows']
        rws = []
        if vars.get_info(eqName)['size'] == vars.get_info(eqName)['nrows']:
            for rr in rows:
                number = vars.initialVals[vars.get_index(eqName, row=rr)]
                rws.append(number)
            basevalues = pd.DataFrame(rws, index=rows)

        else:
            cols = vars.get_info(eqName)['cols']
            for rr in rows:
                clms = []
                for cc in cols:
                    number = vars.initialVals[vars.get_index(eqName, row=rr, col=cc)]
                    clms.append(number)
                rws.append(clms)
            basevalues = pd.DataFrame(rws, index=rows, columns=cols)

    return basevalues


# Get the new values function
def new_value(vars, soln, eqName, ittr):
    #    for scalar
    if vars.get_info(eqName)['size'] == 1:
        newvalue = soln[ittr][vars.get_index(eqName)]

    else:
        rows = vars.get_info(eqName)['rows']
        rws = []
        #       vectors
        if vars.get_info(eqName)['size'] == vars.get_info(eqName)['nrows']:
            for rr in rows:
                number = soln[ittr][vars.get_index(eqName, row=rr)]
                rws.append(number)
            newvalue = pd.DataFrame(rws, index=rows)
        #       data frames
        else:
            cols = []
            for rr in rows:
                cols = vars.get_info(eqName)['cols']
                clms = []
                for cc in cols:
                    number = soln[ittr][vars.get_index(eqName, row=rr, col=cc)]
                    clms.append(number)
                rws.append(clms)
            newvalue = pd.DataFrame(rws, index=rows, columns=cols)

    return newvalue


# Get differences function
def get_diff(vars, soln, eqName, ittr):
    difference = new_value(vars, soln, eqName, ittr) - base_value(vars, eqName)
    return difference



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import json
from pyincore.models.dfr3curve import DFR3Curve


class RestorationCurveSet:
    """class for restoration curves.

    Args:
        metadata (dict): restoration curve metadata.

    Raises:
        ValueError: Raised if there are unsupported number of restoration curves
        or if missing a key curve field.
    """

    def __init__(self, metadata):
        self.id = metadata["id"] if "id" in metadata else ""
        self.description = metadata['description'] if "description" in metadata else ""
        self.authors = ", ".join(metadata['authors']) if "authors" in metadata else ""
        self.paper_reference = str(metadata["paperReference"]) if "paperReference" in metadata else ""
        self.creator = metadata["creator"] if "creator" in metadata else ""
        self.time_units = metadata["timeUnits"]
        self.result_type = metadata["resultType"]
        self.result_unit = metadata["resultUnit"] if "resultUnit" in metadata else ""
        self.hazard_type = metadata['hazardType']
        self.inventory_type = metadata['inventoryType']
        self.restoration_curves = []

        if 'curveParameters' in metadata.keys():
            self.curve_parameters = metadata["curveParameters"]

        if 'restorationCurves' in metadata.keys():
            for restoration_curve in metadata["restorationCurves"]:
                self.restoration_curves.append(DFR3Curve(restoration_curve))
        else:
            raise ValueError("Cannot create dfr3 curve object. Missing key field: restorationCurves.")

    @classmethod
    def from_json_str(cls, json_str):
        """Get dfr3set from json string.

        Args:
            json_str (str): JSON of the Dataset.

        Returns:
            obj: dfr3set from JSON.

        """
        return cls(json.loads(json_str))

    @classmethod
    def from_json_file(cls, file_path):
        """Get dfr3set from the file.

        Args:
            file_path (str): json file path that holds the definition of a dfr3 curve.

        Returns:
            obj: dfr3set from file.

        """
        with open(file_path, "r") as f:
            instance = cls(json.load(f))

        return instance

    def calculate_restoration_rates(self, **kwargs):
        """Computation of restoration rates.

        Args:
            **kwargs: Keyword arguments.

        Returns:
            OrderedDict: Limit state specific restoration rates.

        """

        output = {}

        if len(self.restoration_curves) <= 5:
            for restoration_curve in self.restoration_curves:
                eval_value = restoration_curve.solve_curve_expression(hazard_values={},
                                                                      curve_parameters=self.curve_parameters, **kwargs)
                output[restoration_curve.return_type['description']] = eval_value
        else:
            raise ValueError("We can only handle restoration curves with less than 5 damage states.")

        return output

    def calculate_inverse_restoration_rates(self, **kwargs):
        """Computation of inverse restoration rates example, inverse of cdf, that is, ppf.

        Args:
            **kwargs: Keyword arguments.

        Returns:
            OrderedDict: Limit state specific restoration rates.

        """

        output = {}
        if len(self.restoration_curves) <= 5:
            for restoration_curve in self.restoration_curves:
                eval_value = restoration_curve.solve_curve_for_inverse(hazard_values={},
                                                                       curve_parameters=self.curve_parameters, **kwargs)
                output[restoration_curve.return_type['description']] = eval_value
        else:
            raise ValueError("We can only handle restoration curves with less than 5 damage states.")

        return output



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import concurrent.futures
from itertools import repeat

from pyincore import BaseAnalysis, HazardService, \
    FragilityService, AnalysisUtil, GeoUtil
from pyincore.analyses.buildingdamage.buildingutil import BuildingUtil
from pyincore.models.dfr3curve import DFR3Curve


class BuildingDamage(BaseAnalysis):
    """Building Damage Analysis calculates the probability of building damage based on
    different hazard type such as earthquake, tsunami, and tornado.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        self.hazardsvc = HazardService(incore_client)
        self.fragilitysvc = FragilityService(incore_client)

        super(BuildingDamage, self).__init__(incore_client)

    def run(self):
        """Executes building damage analysis."""
        # Building dataset
        bldg_set = self.get_input_dataset("buildings").get_inventory_reader()

        # building retrofit strategy
        retrofit_strategy_dataset = self.get_input_dataset("retrofit_strategy")
        if retrofit_strategy_dataset is not None:
            retrofit_strategy = list(retrofit_strategy_dataset.get_csv_reader())
        else:
            retrofit_strategy = None

        # Accommodating to multi-hazard
        hazards = []  # hazard objects
        hazard_object = self.get_input_hazard("hazard")

        # To use local hazard
        if hazard_object is not None:
            # Right now only supports single hazard for local hazard object
            hazard_types = [hazard_object.hazard_type]
            hazard_dataset_ids = [hazard_object.id]
            hazards = [hazard_object]
        # To use remote hazard
        elif self.get_parameter("hazard_id") is not None and self.get_parameter("hazard_type") is not None:
            hazard_dataset_ids = self.get_parameter("hazard_id").split("+")
            hazard_types = self.get_parameter("hazard_type").split("+")
            for hazard_type, hazard_dataset_id in zip(hazard_types, hazard_dataset_ids):
                hazards.append(BaseAnalysis._create_hazard_object(hazard_type, hazard_dataset_id, self.hazardsvc))
        else:
            raise ValueError("Either hazard object or hazard id + hazard type must be provided")

        # Get Fragility key
        fragility_key = self.get_parameter("fragility_key")
        if fragility_key is None:
            fragility_key = BuildingUtil.DEFAULT_TSUNAMI_MMAX_FRAGILITY_KEY if 'tsunami' in hazard_types else \
                    BuildingUtil.DEFAULT_FRAGILITY_KEY
            self.set_parameter("fragility_key", fragility_key)

        user_defined_cpu = 1

        if not self.get_parameter("num_cpu") is None and self.get_parameter("num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self, len(bldg_set), user_defined_cpu)

        avg_bulk_input_size = int(len(bldg_set) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(bldg_set)
        while count < len(inventory_list):
            inventory_args.append(inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        (ds_results, damage_results) = self.building_damage_concurrent_future(self.building_damage_analysis_bulk_input,
                                                                              num_workers,
                                                                              inventory_args,
                                                                              repeat(retrofit_strategy),
                                                                              repeat(hazards),
                                                                              repeat(hazard_types),
                                                                              repeat(hazard_dataset_ids))

        self.set_result_csv_data("ds_result", ds_results, name=self.get_parameter("result_name"))
        self.set_result_json_data("damage_result",
                                  damage_results,
                                  name=self.get_parameter("result_name") + "_additional_info")

        return True

    def building_damage_concurrent_future(self, function_name, parallelism, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            parallelism (int): Number of workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of ordered dictionaries with building damage values and other data/metadata.

        """
        output_ds = []
        output_dmg = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=parallelism) as executor:
            for ret1, ret2 in executor.map(function_name, *args):
                output_ds.extend(ret1)
                output_dmg.extend(ret2)

        return output_ds, output_dmg

    def building_damage_analysis_bulk_input(self, buildings, retrofit_strategy, hazards, hazard_types,
                                            hazard_dataset_ids):
        """Run analysis for multiple buildings.

        Args:
            buildings (list): Multiple buildings from input inventory set.
            retrofit_strategy (list): building guid and its retrofit level 0, 1, 2, etc. This is Optional
            hazards (list): List of hazard objects.
            hazard_types (list): List of Hazard type, either earthquake, tornado, or tsunami.
            hazard_dataset_ids (list): List of id of the hazard exposure.

        Returns:
            list: A list of ordered dictionaries with building damage values and other data/metadata.

        """

        fragility_key = self.get_parameter("fragility_key")
        fragility_sets = self.fragilitysvc.match_inventory(self.get_input_dataset("dfr3_mapping_set"), buildings,
                                                           fragility_key, retrofit_strategy)
        use_liquefaction = False
        liquefaction_resp = None
        # Get geology dataset id containing liquefaction susceptibility
        geology_dataset_id = self.get_parameter("liquefaction_geology_dataset_id")

        multihazard_vals = {}
        adjust_demand_types_mapping = {}

        for hazard, hazard_type, hazard_dataset_id in zip(hazards, hazard_types, hazard_dataset_ids):
            # get allowed demand types for the hazard type
            allowed_demand_types = [item["demand_type"].lower() for item in self.hazardsvc.get_allowed_demands(
                hazard_type)]

            # Liquefaction
            if hazard_type == "earthquake" and self.get_parameter("use_liquefaction") is not None:
                use_liquefaction = self.get_parameter("use_liquefaction")

            values_payload = []
            values_payload_liq = []  # for liquefaction, if used
            unmapped_buildings = []
            mapped_buildings = []
            for b in buildings:
                bldg_id = b["id"]
                if bldg_id in fragility_sets:
                    location = GeoUtil.get_location(b)
                    loc = str(location.y) + "," + str(location.x)
                    demands, units, adjusted_to_original = \
                        AnalysisUtil.get_hazard_demand_types_units(b,
                                                                   fragility_sets[bldg_id],
                                                                   hazard_type,
                                                                   allowed_demand_types)
                    adjust_demand_types_mapping.update(adjusted_to_original)
                    value = {
                        "demands": demands,
                        "units": units,
                        "loc": loc
                    }
                    values_payload.append(value)
                    mapped_buildings.append(b)

                    if use_liquefaction and geology_dataset_id is not None:
                        value_liq = {
                            "demands": [""],
                            "units": [""],
                            "loc": loc
                        }
                        values_payload_liq.append(value_liq)
                else:
                    unmapped_buildings.append(b)

            hazard_vals = hazard.read_hazard_values(values_payload, self.hazardsvc)

            # map demand type from payload to response
            # worst code I have ever written
            # e.g. 1.04 Sec Sa --> 1.04 SA --> 1.0 SA
            for payload, response in zip(values_payload, hazard_vals):
                for i in range(len(payload["demands"])):
                    adjust_demand_types_mapping[response["demands"][i]] = adjust_demand_types_mapping[payload[
                        "demands"][i]]

            # record hazard value for each hazard type for later calcu
            multihazard_vals[hazard_type] = hazard_vals

            # Check if liquefaction is applicable
            if hazard_type == "earthquake" and use_liquefaction and geology_dataset_id is not None:
                liquefaction_resp = self.hazardsvc.post_liquefaction_values(hazard_dataset_id, geology_dataset_id,
                                                                            values_payload_liq)

        # not needed anymore as they are already split into mapped and unmapped
        del buildings

        ds_results = []
        damage_results = []

        i = 0
        for b in mapped_buildings:
            ds_result = dict()
            damage_result = dict()
            dmg_probability = dict()
            dmg_interval = dict()
            b_id = b["id"]
            selected_fragility_set = fragility_sets[b_id]
            ground_failure_prob = None

            # TODO: Once all fragilities are migrated to new format, we can remove this condition
            if isinstance(selected_fragility_set.fragility_curves[0], DFR3Curve):
                # Supports multiple hazard and multiple demand types in same fragility
                hval_dict = dict()
                b_multihaz_vals = dict()
                b_demands = dict()
                b_units = dict()
                for hazard_type in hazard_types:
                    b_haz_vals = AnalysisUtil.update_precision_of_lists(multihazard_vals[hazard_type][i][
                                                                            "hazardValues"])
                    b_demands[hazard_type] = multihazard_vals[hazard_type][i]["demands"]
                    b_units[hazard_type] = multihazard_vals[hazard_type][i]["units"]
                    b_multihaz_vals[hazard_type] = b_haz_vals
                    # To calculate damage, use demand type name from fragility that will be used in the expression,
                    # instead  of using what the hazard service returns. There could be a difference "SA" in DFR3 vs
                    # "1.07 SA" from hazard
                    j = 0
                    for adjusted_demand_type in multihazard_vals[hazard_type][i]["demands"]:
                        d = adjust_demand_types_mapping[adjusted_demand_type]
                        hval_dict[d] = b_haz_vals[j]
                        j += 1

                # catch any of the hazard values error
                hazard_values_errors = False
                for hazard_type in hazard_types:
                    hazard_values_errors = hazard_values_errors or AnalysisUtil.do_hazard_values_have_errors(
                        b_multihaz_vals[hazard_type])

                if not hazard_values_errors:
                    building_args = selected_fragility_set.construct_expression_args_from_inventory(b)

                    building_period = selected_fragility_set.fragility_curves[0].get_building_period(
                        selected_fragility_set.curve_parameters, **building_args)

                    dmg_probability = selected_fragility_set.calculate_limit_state(
                        hval_dict, **building_args, period=building_period)

                    if use_liquefaction and geology_dataset_id is not None and liquefaction_resp is not None:
                        ground_failure_prob = liquefaction_resp[i][BuildingUtil.GROUND_FAILURE_PROB]
                        dmg_probability = AnalysisUtil.update_precision_of_dicts(
                            AnalysisUtil.adjust_damage_for_liquefaction(dmg_probability, ground_failure_prob))

                    dmg_interval = selected_fragility_set.calculate_damage_interval(
                        dmg_probability, hazard_type="+".join(hazard_types), inventory_type="building")
            else:
                raise ValueError("One of the fragilities is in deprecated format. This should not happen. If you are "
                                 "seeing this please report the issue.")

            ds_result['guid'] = b['properties']['guid']
            damage_result['guid'] = b['properties']['guid']

            ds_result.update(dmg_probability)
            ds_result.update(dmg_interval)

            # determine expose from multiple hazard
            haz_expose = False
            for hazard_type in hazard_types:
                haz_expose = haz_expose or AnalysisUtil.get_exposure_from_hazard_values(b_multihaz_vals[
                                                                                             hazard_type], hazard_type)
            ds_result['haz_expose'] = haz_expose

            damage_result['fragility_id'] = selected_fragility_set.id
            damage_result['demandtype'] = b_demands
            damage_result['demandunits'] = b_units
            damage_result['hazardval'] = b_multihaz_vals

            if use_liquefaction and geology_dataset_id is not None:
                damage_result[BuildingUtil.GROUND_FAILURE_PROB] = ground_failure_prob

            ds_results.append(ds_result)
            damage_results.append(damage_result)
            i += 1

        for b in unmapped_buildings:
            ds_result = dict()
            damage_result = dict()
            ds_result['guid'] = b['properties']['guid']
            damage_result['guid'] = b['properties']['guid']
            damage_result['fragility_id'] = None
            damage_result['demandtype'] = None
            damage_result['demandunits'] = None
            damage_result['hazardval'] = None

            ds_results.append(ds_result)
            damage_results.append(damage_result)

        return ds_results, damage_results

    def get_spec(self):
        """Get specifications of the building damage analysis.

        Returns:
            obj: A JSON object of specifications of the building damage analysis.

        """
        return {
            'name': 'building-damage',
            'description': 'building damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'hazard_type',
                    'required': False,
                    'description': 'Hazard Type (e.g. earthquake)',
                    'type': str
                },
                {
                    'id': 'hazard_id',
                    'required': False,
                    'description': 'Hazard ID',
                    'type': str
                },
                {
                    'id': 'fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'use_liquefaction',
                    'required': False,
                    'description': 'Use liquefaction',
                    'type': bool
                },
                {
                    'id': 'use_hazard_uncertainty',
                    'required': False,
                    'description': 'Use hazard uncertainty',
                    'type': bool
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
                {
                    'id': 'seed',
                    'required': False,
                    'description': 'Initial seed for the tornado hazard value',
                    'type': int
                },
                {
                    'id': 'liquefaction_geology_dataset_id',
                    'required': False,
                    'description': 'Geology dataset id',
                    'type': str,
                }
            ],
            'input_hazards': [
                {
                    'id': 'hazard',
                    'required': False,
                    'description': 'Hazard object',
                    'type': ["earthquake", "tornado", "hurricane", "flood", "tsunami"]
                },
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventoryVer4', 'ergo:buildingInventoryVer5',
                             'ergo:buildingInventoryVer6', 'ergo:buildingInventoryVer7'],
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                },
                {
                    'id': 'retrofit_strategy',
                    'required': False,
                    'description': 'Building retrofit strategy that contains guid and retrofit method',
                    'type': ['incore:retrofitStrategy']
                }
            ],
            'output_datasets': [
                {
                    'id': 'ds_result',
                    'parent_type': 'buildings',
                    'description': 'CSV file of damage states for building structural damage',
                    'type': 'ergo:buildingDamageVer6'
                },
                {
                    'id': 'damage_result',
                    'parent_type': 'buildings',
                    'description': 'Json file with information about applied hazard value and fragility',
                    'type': 'incore:buildingDamageSupplement'
                }
            ]
        }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
from pyincore import HazardService, Dataset
from pyincore.models.hazard.hazard import Hazard
from pyincore.models.hazard.hazarddataset import TornadoDataset
from pyincore.models.units import Units

from shapely.geometry import Point
import random
import time
import math


class Tornado(Hazard):
    """
    "tornadoType": "dataset",
    "id": "xxx",
    "name": "pytest - Joplin Tornado",
    "description": "Joplin tornado hazard",
    "creator": "xxx",
    "threshold": null,
    "thresholdUnit": "mph",
    "spaces": [
        "xxx"
    ],
    "date": "2020-08-14T16:22:32+0000",
    "datasetId": "xxx"
    """

    def __init__(self, metadata, ef_rating_field="ef_rating", ef_wind_speed=(65, 86, 111, 136, 166, 200),
                 max_wind_speed=250.0):
        super().__init__(metadata)
        self.tornado_type = metadata["tornadoType"] if "tornadoType" in metadata else ""
        # tornado has very different shape than other hazards
        if self.tornado_type == "dataset":
            self.hazardDatasets = [
                TornadoDataset({"threshold": metadata["threshold"] if "threshold" in metadata else None,
                                "demandType": "wind",
                                "demandUnits": metadata["thresholdUnit"] if "thresholdUnit" in metadata else "mph",
                                "datasetId": metadata["datasetId"] if "datasetId" in metadata else ""})
            ]
        else:
            self.hazardDatasets = []
        self.hazard_type = "tornado"
        self.EF_RATING_FIELD = ef_rating_field
        self.EF_WIND_SPEED = ef_wind_speed
        self.MAX_WIND_SPEED = max_wind_speed
        self.tornado_parameters = metadata["TornadoParameters"] if "TornadoParameters" in metadata else {}

    @classmethod
    def from_hazard_service(cls, id: str, hazard_service: HazardService):
        """Get Hazard from hazard service, get metadata as well.

        Args:
            id (str): ID of the Hazard.
            hazard_service (obj): Hazard service.

        Returns:
            obj: Hazard from Data service.

        """
        metadata = hazard_service.get_tornado_hazard_metadata(id)
        instance = cls(metadata)
        return instance

    def read_hazard_values(self, payload: list, hazard_service=None, seed=None, **kwargs):
        """ Retrieve bulk earthquake hazard values either from the Hazard service or read it from local Dataset

        Args:
            payload (list):
            seed: (None or int): Seed value for random values.
            hazard_service (obj): Hazard service.
            kwargs (dict): Keyword arguments.
        Returns:
            obj: Hazard values.

        """
        if self.id and self.id != "" and hazard_service is not None:
            return hazard_service.post_tornado_hazard_values(self.id, payload, seed=seed, **kwargs)
        else:
            if self.tornado_type == "dataset":
                return self.calculate_wind_speed_uniform_random_dist(payload, seed)
            else:
                raise ValueError("Local Tornado type \"" + self.tornado_type + "\" is not supported yet.")

    def calculate_wind_speed_uniform_random_dist(self, payload, seed=-1):
        """ Read local hazard values from shapefile dataset

                Args:
                    payload (list):
                    seed: (None or int): Seed value for random values.
                Returns:
                    obj: Hazard values.

                """

        response = []

        # match demand types with shapefile file
        for req in payload:
            hazard_values = []
            for index, req_demand_type in enumerate(req["demands"]):
                match = False
                for hazard_dataset in self.hazardDatasets:
                    if hazard_dataset.dataset is None or not isinstance(hazard_dataset.dataset, Dataset):
                        raise Exception("Hazard dataset is not properly attached to the hazard object.")

                    # find matching raster file (Dataset) to read value from
                    if req_demand_type.lower() == hazard_dataset.demand_type.lower():
                        hazard_df = hazard_dataset.dataset.get_dataframe_from_shapefile()
                        ef_box = -1
                        x = float(req["loc"].split(",")[1])
                        y = float(req["loc"].split(",")[0])
                        location = Point(x, y)

                        for _, feature in hazard_df.iterrows():
                            polygon = feature['geometry']
                            if location.within(polygon):
                                ef_rating = feature[self.EF_RATING_FIELD]
                                ef_box = Tornado.get_ef_rating(ef_rating)
                                break

                        if ef_box < 0:
                            raw_wind_speed = None
                        elif ef_box == 5:  # EF5
                            bottom_speed = self.EF_WIND_SPEED[ef_box]
                            top_speed = self.MAX_WIND_SPEED
                            random.seed(self.get_random_seed(location, seed))
                            raw_wind_speed = random.uniform(bottom_speed, top_speed)
                        else:
                            bottom_speed = self.EF_WIND_SPEED[ef_box]
                            top_speed = self.EF_WIND_SPEED[ef_box + 1]
                            random.seed(self.get_random_seed(location, seed))
                            raw_wind_speed = random.uniform(bottom_speed, top_speed)

                        if raw_wind_speed is None:
                            converted_wind_speed = raw_wind_speed
                        else:
                            # some basic unit conversion
                            converted_wind_speed = Units.convert_hazard(raw_wind_speed,
                                                                        original_demand_units=hazard_dataset.demand_units,
                                                                        requested_demand_units=req["units"][index])

                            # compare with threshold (optional)
                            threshold_value = hazard_dataset.threshold_value
                            threshold_unit = hazard_dataset.threshold_unit
                            if threshold_value is not None:
                                converted_threshold_value = Units.convert_hazard(threshold_value,
                                                                                 original_demand_units=threshold_unit,
                                                                                 requested_demand_units=req["units"][
                                                                                     index])
                                if converted_wind_speed < converted_threshold_value:
                                    converted_wind_speed = None

                        hazard_values.append(converted_wind_speed)
                        match = True
                        break

                if not match:
                    hazard_values.append(-9999.2)  # invalid demand type

            req.update({"hazardValues": hazard_values})
            response.append(req)

        return response

    def get_random_seed(self, location, seed=-1):

        # Get seed from the model and override if no value specified
        if (seed is None or seed == -1) and self.tornado_parameters is not {} and "randomSeed" in self.tornado_parameters:
            seed = self.tornado_parameters["randomSeed"]

        # If no seed value provided OR model seed value was never set by the user, use current system time
        if seed is None or seed == -1:
            seed = int(time.time() * 1000)  # Current system time in milliseconds

        # Use 4 decimal places for getting unique seed values from lat/long
        try:
            seed = seed + int(abs((location.x + location.y) * 10000))
        except OverflowError:
            print("Seed + abs((location.x + location.y) * 10000) exceeds max value, capping at Maximum value")
            seed = float('inf')  # Cap at positive infinity for maximum value

        return seed

    @staticmethod
    def get_ef_rating(ef_rating):
        ef_rating_map = {
            "EF0": 0,
            "EF1": 1,
            "EF2": 2,
            "EF3": 3,
            "EF4": 4,
            "EF5": 5,
        }
        return ef_rating_map.get(ef_rating, -1)



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

from pyincore.analyses.indp.indpcomponents import INDPComponents
import os


class INDPResults:
    """
    This class saves INDP results including actions, costs, run time, and components

    Attributes
    ----------
    results : dict
        Dictionary containing INDP results including actions, costs, run time, and components
    layers : list
        List of layers in the analysis
    results_layer : int
        Dictionary containing INDP results for each layer including actions, costs, run time, and components
    """
    cost_types = ["Space Prep", "Arc", "Node", "Over Supply", "Under Supply", "Flow", "Total", "Under Supply Perc"]

    def __init__(self, layers=None):
        if layers is None:
            layers = []
        self.results = {}
        self.layers = layers
        self.results_layer = {layer: {} for layer in layers}

    def __len__(self):
        return len(self.results)

    def __getitem__(self, index):
        return self.results[index]

    def extend(self, indp_result, t_offset=0, t_start=0, t_end=0):
        """
        This function extends the results to accommodate a new time step.

        Parameters
        ----------
        indp_result : :class:`~INDPResults`
            The current :class:`~INDPResults` object before extension
        t_offset : int
            The number of time steps that the current results should be shifted forward. The default is 0.
        t_start : int
            The starting time step. The default is 0.
        t_end : int
            The ending time step. The default is 0.

        Returns
        -------
        None.

        """
        if t_end == 0:
            t_end = len(indp_result)
        for new_t, t in zip([x + t_offset for x in range(t_end - t_start)],
                            [y + t_start for y in range(t_end - t_start)]):
            self.results[new_t] = indp_result.results[t]
        if self.layers:
            if t_end == 0:
                t_end = len(indp_result[self.layers[0]])
            for layer in indp_result.results_layer.keys():
                for new_t, t in zip([x + t_offset for x in range(t_end - t_start)],
                                    [y + t_start for y in range(t_end - t_start)]):
                    self.results_layer[layer][new_t] = indp_result.results_layer[layer][t]

    def add_cost(self, t, cost_type, cost, cost_layer=None):
        """
        This function adds cost values to the results.

        Parameters
        ----------
        t : int
            The time steps to which the costs should be added
        cost_type : str
            The cost types that is added. The options are: "Space Prep", "Arc", "Node", "Over Supply", "Under Supply",
            "Flow", "Total", "Under Supply Perc"
        cost : float
            The cost value that is added
        cost_layer : dict
            The cost value that is added for each layer. The default is None, which adds no value for layers.

        Returns
        -------
        None.

        """
        if cost_layer is None:
            cost_layer = {}
        if t not in self.results:
            self.results[t] = {
                'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                          "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
        self.results[t]['costs'][cost_type] = cost
        if self.layers:
            for layer in cost_layer.keys():
                if t not in self.results_layer[layer]:
                    self.results_layer[layer][t] = {
                        'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                                  "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                        'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
                self.results_layer[layer][t]['costs'][cost_type] = cost_layer[layer]

    def add_run_time(self, t, run_time, save_layer=True):
        """
        This function adds run time to the results.

        Parameters
        ----------
        t : int
            The time steps to which the run time should be added
        run_time : float
            The run time that is added
        save_layer : bool
            If the run time should be added for each layer. The default is True.

        Returns
        -------
        None.

        """
        if t not in self.results:
            self.results[t] = {
                'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                          "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
        self.results[t]['run_time'] = run_time
        if self.layers and save_layer:
            for layer in self.layers:
                if t not in self.results_layer[layer]:
                    self.results_layer[layer][t] = {
                        'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                                  "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                        'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
                self.results_layer[layer][t]['run_time'] = run_time

    def add_action(self, t, action, save_layer=True):
        """
        This function adds restoration actions to the results.

        Parameters
        ----------
        t : int
            The time steps to which the actions should be added
        action : list
            List of actions that are added
        save_layer : bool
            If the actions should be added for each layer. The default is True.

        Returns
        -------
        None.

        """
        if t not in self.results:
            self.results[t] = {
                'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                          "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
        self.results[t]['actions'].append(action)
        if self.layers and save_layer:
            action_layer = int(action[-1])
            if t not in self.results_layer[action_layer]:
                self.results_layer[action_layer][t] = {
                    'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                              "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                    'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
            self.results_layer[action_layer][t]['actions'].append(action)

    def add_gc_size(self, t, gc_size):
        """
        This function adds the giant component size to the results

        Parameters
        ----------
        t : int
            The time steps to which the giant component size should be added
        gc_size : int
            The giant component size that is added

        Returns
        -------
        None.

        """
        if t not in self.results:
            self.results[t] = {
                'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                          "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
        self.results[t]['gc_size'] = gc_size

    def add_num_components(self, t, num_components):
        """
        This function adds the number of components to the results

        Parameters
        ----------
        t : int
            The time steps to which the number of components should be added
        num_components : int
            The number of components that is added

        Returns
        -------
        None.

        """
        if t not in self.results:
            self.results[t] = {
                'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                          "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
        self.results[t]['num_components'] = num_components

    def add_components(self, t, components):
        """
        This function adds the components to the results

        Parameters
        ----------
        t : int
            The time steps to which the number of components should be added
        components : list
            The list of components that is added

        Returns
        -------
        None.

        """
        if t not in self.results:
            self.results[t] = {
                'costs': {"Space Prep": 0.0, "Arc": 0.0, "Node": 0.0, "Over Supply": 0.0, "Under Supply": 0.0,
                          "Under Supply Perc": 0.0, "Flow": 0.0, "Total": 0.0}, 'actions': [], 'gc_size': 0,
                'num_components': 0, 'components': INDPComponents(), 'run_time': 0.0}
        self.results[t]['components'] = components
        self.add_num_components(t, components.num_components)
        self.add_gc_size(t, components.gc_size)

    def to_csv_layer(self, out_dir, sample_num=1, suffix=""):
        """
        This function writes the results to file for each layer. The file for each layer are distinguished by "_L" +
        the layer number.

        Parameters
        ----------
        out_dir : str
            Output directory to which the results should be written.
        sample_num : int
            The sample number corresponding to the results, The default is 1.
        suffix : str
            The suffix of the file that is being written. The default is "".

        Returns
        -------
        None.

        """
        for layer in self.layers:
            action_file = out_dir + "/actions_" + str(sample_num) + "_L" + str(layer) + "_" + suffix + ".csv"
            costs_file = out_dir + "/costs_" + str(sample_num) + "_L" + str(layer) + "_" + suffix + ".csv"
            run_time_file = out_dir + "/run_time_" + str(sample_num) + "_L" + str(layer) + "_" + suffix + ".csv"
            with open(action_file, 'w') as f:
                f.write("t,action\n")
                for t in self.results_layer[layer]:
                    for a in self.results_layer[layer][t]['actions']:
                        f.write(str(t) + "," + a + "\n")
            with open(run_time_file, 'w') as f:
                f.write("t,run_time\n")
                for t in self.results_layer[layer]:
                    f.write(str(t) + "," + str(self.results_layer[layer][t]['run_time']) + "\n")
            with open(costs_file, 'w') as f:
                f.write("t,Space Prep,Arc,Node,Over Supply,Under Supply,Flow,Total,Under Supply Perc\n")
                for t in self.results_layer[layer]:
                    costs = self.results_layer[layer][t]['costs']
                    f.write(str(t) + "," + str(costs["Space Prep"]) + "," + str(costs["Arc"]) + "," + str(
                        costs["Node"]) + "," + str(costs["Over Supply"]) + "," + str(costs["Under Supply"]) + "," + str(
                        costs["Flow"]) + "," + str(costs["Total"]) + "," + str(costs["Under Supply Perc"]) + "\n")

    @classmethod
    def from_csv(clss, out_dir, sample_num=1, suffix=""):
        """
        This function reads the results from file.

        Parameters
        ----------
        out_dir : str
            Output directory from which the results should be read
        sample_num : int
            The sample number corresponding to the results, The default is 1.
        suffix : str
            The suffix of the file that is being read. The default is "".

        Returns
        -------
        indp_result: :class:`~INDPResults`
            The :class:`~INDPResults` object containing the read results

        """
        action_file = out_dir + "/actions_" + str(sample_num) + "_" + suffix + ".csv"
        costs_file = out_dir + "/costs_" + str(sample_num) + "_" + suffix + ".csv"
        perc_file = out_dir + "/percolation_" + str(sample_num) + "_" + suffix + ".csv"
        comp_file = out_dir + "/components_" + str(sample_num) + "_" + suffix + ".csv"
        run_time_file = out_dir + "/run_time_" + str(sample_num) + "_" + suffix + ".csv"
        indp_result = INDPResults()
        if os.path.isfile(action_file):  # ..todo: component-related results are not currently added to the results.
            with open(action_file) as f:
                lines = f.readlines()[1:]
                for line in lines:
                    data = line.strip().split(',')
                    t = int(data[0])
                    action = str.strip(data[1])
                    indp_result.add_action(t, action)
            with open(costs_file) as f:
                lines = f.readlines()
                cost_types = lines[0].strip().split(',')[1:]
                for line in lines[1:]:
                    data = line.strip().split(',')
                    t = int(data[0])
                    costs = data[1:]
                    for ct in range(len(cost_types)):
                        indp_result.add_cost(t, cost_types[ct], float(costs[ct]))
            with open(run_time_file) as f:
                lines = f.readlines()
                for line in lines[1:]:
                    data = line.strip().split(',')
                    t = int(data[0])
                    run_time = data[1]
                    indp_result.add_run_time(t, run_time)
            # with open(perc_file) as f:
            #    lines=f.readlines()[1:]
            #    for line in lines:
            #        data=string.split(str.strip(line),",")
            #        t=int(data[0])
            #        indp_result.add_gc_size(t,int(data[1]))
            #        indp_result.add_num_components(t,int(data[2]))
            # with open(comp_file) as f:
            #    lines=f.readlines()[1:]
            #    for line in lines:
            #        data=string.split(str.strip(line),",")
            #        t=int(data[0])
            #        comps=data[1:]
            #        if comps[0]!='':
            #            indp_result.add_components(t,INDPComponents.from_csv_string(comps))
            #        else:
            #            print("Caution: No component.")
        else:
            raise ValueError('File does not exist: ' + action_file)
        return indp_result



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import sys, random


class Solution:
    """Abstract solution. To be implemented."""

    def __init__(self, num_objectives):
        """Constructor. Parameters: number of objectives."""
        self.num_objectives = num_objectives
        self.objectives = []
        for _ in range(num_objectives):
            self.objectives.append(None)
        self.attributes = []
        self.rank = sys.maxsize
        self.distance = 0.0

        self.chromos_fitness = {}
        self.sch = {}

    def __rshift__(self, other):
        """True if this solution dominates the other (">>" operator)."""
        dominates = False

        for i in range(len(self.objectives)):
            if self.objectives[i] > other.objectives[i]:
                return False

            elif self.objectives[i] < other.objectives[i]:
                dominates = True

        return dominates

    def __lshift__(self, other):
        """True if this solution is dominated by the other ("<<" operator)."""
        return other >> self


def crowded_comparison(s1, s2):
    """Compare the two solutions based on crowded comparison.

    Args:
        s1 (obj): One chromosome.
        s2 (obj): Another chromosome.

    Returns:
        float: A comparison value.

    """
    if s1.rank < s2.rank:
        return 1

    elif s1.rank > s2.rank:
        return -1

    elif s1.distance > s2.distance:
        return 1

    elif s1.distance < s2.distance:
        return -1

    else:
        return 0


class NSGAII:
    """Implementation of NSGA-II algorithm."""
    current_evaluated_objective = 0

    def __init__(self, num_objectives, mutation_rate=0.1, crossover_rate=1.0):
        """Constructor.

        Args:
            num_objectives (obj): Number of objectives.
            mutation_rate (float): Mutation rate (default value 10%).
            crossover_rate (float): Crossover rate (default value 100%)..

        """
        self.num_objectives = num_objectives
        self.mutation_rate = mutation_rate
        self.crossover_rate = crossover_rate

        random.seed(100)

    def run(self, p, population_size, num_generations):
        """Run NSGA-II.

        Args:
            p (obj): A set of chromosomes (population).
            population_size (obj): A population size.
            num_generations (obj): A number of generations.

        Returns:
            list: First front of Pareto front.

        """
        for s in p:
            s.evaluate_solution(0)

        first_front = []
        for i in range(num_generations):

            r = []
            r.extend(p)

            fronts = self.fast_nondominated_sort(r)

            del p[:]

            for front in fronts.values():
                if len(front) == 0:
                    break

                self.crowding_distance_assignment(front)
                p.extend(front)

                if len(p) >= population_size:
                    break

            self.sort_crowding(p)

            if len(p) > population_size:
                del p[population_size:]

            first_front = list(fronts.values())[0]

        return first_front

    @staticmethod
    def sort_ranking(p):
        """Run sort the sort of chromosomes according to their ranks.

        Args:
            p (obj): A set of chromosomes (population).

        """
        for i in range(len(p) - 1, -1, -1):
            for j in range(1, i + 1):
                s1 = p[j - 1]
                s2 = p[j]

                if s1.rank > s2.rank:
                    p[j - 1] = s2
                    p[j] = s1

    @staticmethod
    def sort_objective(p, obj_idx):
        """Run sort the chromosome based on their objective value.

        Args:
            p (obj): A set of chromosomes (population).
            obj_idx (int): The index of objective function.

        """
        for i in range(len(p) - 1, -1, -1):
            for j in range(1, i + 1):
                s1 = p[j - 1]
                s2 = p[j]

                if s1.objectives[obj_idx] > s2.objectives[obj_idx]:
                    p[j - 1] = s2
                    p[j] = s1

    @staticmethod
    def sort_crowding(p):
        """Run calculate the crowding distance of adjacent two chromosome in a front level.

        Args:
            p (obj): A set of chromosomes (population).

        """
        for i in range(len(p) - 1, -1, -1):
            for j in range(1, i + 1):
                s1 = p[j - 1]
                s2 = p[j]

                if crowded_comparison(s1, s2) < 0:
                    p[j - 1] = s2
                    p[j] = s1

    def make_new_pop(self, p):
        """Make new population Q, offspring of P.

        Args:
            p (obj): A set of chromosomes (population).

        Returns:
            list: Offspring.

        """
        q = []

        while len(q) != len(p):
            selected_solutions = [None, None]

            while selected_solutions[0] == selected_solutions[1]:
                for i in range(2):
                    s1 = random.choice(p)
                    s2 = s1
                    while s1 == s2:
                        s2 = random.choice(p)

                    if crowded_comparison(s1, s2) > 0:
                        selected_solutions[i] = s1

                    else:
                        selected_solutions[i] = s2

            if random.random() < self.crossover_rate:
                child_solution = selected_solutions[0].crossover(selected_solutions[1])

                if random.random() < self.mutation_rate:
                    child_solution.mutate()

                child_solution.evaluate_solution(0)

                q.append(child_solution)

        return q

    @staticmethod
    def fast_nondominated_sort(p):
        """Discover Pareto fronts in P, based on non-domination criterion.

        Args:
            p (obj): A set of chromosomes (population).

        Returns:
            dict: Fronts.

        """
        fronts = {}

        s = {}
        n = {}
        for i in p:
            s[i] = []
            n[i] = 0

        fronts[1] = []

        for pk in p:
            for qk in p:
                if pk == qk:
                    continue

                if pk >> qk:
                    s[pk].append(qk)

                elif pk << qk:
                    n[pk] += 1

            if n[pk] == 0:
                fronts[1].append(pk)

        i = 1
        while len(fronts[i]) != 0:
            next_front = []
            for r in fronts[i]:
                for j in s[r]:
                    n[j] -= 1
                    if n[j] == 0:
                        next_front.append(j)

            i += 1
            fronts[i] = next_front

        return fronts

    def crowding_distance_assignment(self, front):
        """Assign a crowding distance for each solution in the front.

        Args:
            front (dict): A set of chromosomes in the front level.

        """
        for p in front:
            p.distance = 0

        for obj_index in range(self.num_objectives):
            self.sort_objective(front, obj_index)

            front[0].distance = float('inf')
            front[len(front) - 1].distance = float('inf')

            for i in range(1, len(front) - 1):
                front[i].distance += (front[i + 1].distance - front[i - 1].distance)



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class Mapping:
    """ mapping class that contains the rules and keys of dfr3 curves.

    Args:
        entry (dict): mapping entry.
        rules (list): mapping match rules

    """

    def __init__(self, entry: dict, rules: list):
        self.entry = entry
        self.rules = rules



from __future__ import division
import pandas as pd
import copy
import random

from pyincore.analyses.transportationrecovery.transportationrecoveryutil import TransportationRecoveryUtil
from pyincore.analyses.transportationrecovery.nsga2 import NSGAII
from pyincore.analyses.transportationrecovery import WIPW as WIPW
from pyincore.analyses.transportationrecovery.post_disaster_long_term_solution import PostDisasterLongTermSolution
from pyincore import BaseAnalysis


class TransportationRecovery(BaseAnalysis):

    def run(self):
        """ Executes transportation recovery analysis"""

        # read the nodes in transportation
        node_set = self.get_input_dataset("nodes").get_inventory_reader()
        nodes = list(node_set)
        node_df = pd.DataFrame(columns=nodes[0]['properties'].keys())
        for node in nodes:
            node_properties_df = pd.DataFrame.from_dict(node['properties'], orient='index').T
            node_df = pd.concat([node_df, node_properties_df], ignore_index=True)

        # read the link in transportation
        link_set = self.get_input_dataset("links").get_inventory_reader()
        links = list(link_set)
        arc_df = pd.DataFrame(columns=links[0]['properties'].keys())
        for link in links:
            link_properties_df = pd.DataFrame.from_dict(link['properties'], orient='index').T
            arc_df = pd.concat([arc_df, link_properties_df], ignore_index=True)

        # read bridge information
        bridge_set = self.get_input_dataset("bridges").get_inventory_reader()
        bridges = list(bridge_set)
        bridge_df = pd.DataFrame(columns=bridges[0]['properties'].keys())
        for bridge in bridges:
            bridge_properties_df = pd.DataFrame.from_dict(bridge['properties'], orient='index').T
            bridge_df = pd.concat([bridge_df, bridge_properties_df], ignore_index=True)

        # read bridge damage information
        bridge_damage_value = self.get_input_dataset("bridge_damage_value").get_json_reader()
        unrepaired_bridge = self.get_input_dataset("unrepaired_bridge").get_json_reader()

        seed = 333
        random.seed(seed)

        # record the average daily traffic (ADT) data of bridge, a road has the
        # same ADT value as its crossing bridge
        adt_data = self.get_input_dataset("ADT").get_json_reader()
        ADT = copy.deepcopy(adt_data)

        # if there is no bridge across a road, the ADT of the road equals to
        # the maximum value of ADTs
        for i in range(len(arc_df)):
            if arc_df["guid"][i] not in list(adt_data.keys()):
                adt_data[arc_df["guid"][i]] = max(ADT.values())

        pm = self.get_parameter("pm")
        ini_num_population = self.get_parameter("ini_num_population")
        population_size = self.get_parameter("population_size")
        num_generation = self.get_parameter("num_generation")
        mutation_rate = self.get_parameter("mutation_rate")
        crossover_rate = self.get_parameter("crossover_rate")

        # create network
        network = TransportationRecoveryUtil.nw_reconstruct(node_df, arc_df, adt_data)

        if pm == 0:
            # calculate the WIPW (Weighted independent pathways) index
            all_ipw, all_ipw_length = WIPW.ipw_search(network.nodes(), network.edges())
            # Calculate the ADT for each path
            path_adt = {}
            for node in range(len(network.nodes()) - 1):
                for pairnode in range(node + 1, len(network.nodes())):
                    if (node, pairnode) in all_ipw.keys():
                        path_adt[node, pairnode] = {}
                        path_adt[pairnode, node] = {}
                        for key, value in all_ipw[node, pairnode].items():
                            path_adt[node, pairnode][key] \
                                = WIPW.path_adt_from_edges(network, value)
                            path_adt[pairnode, node][key] \
                                = path_adt[node, pairnode][key]
        else:
            all_ipw = None
            all_ipw_length = None
            path_adt = None

        num_objectives = 2

        # implement NSGA for transportation network post-disaster recovery
        nsga2 = NSGAII(num_objectives, mutation_rate, crossover_rate)

        p = []
        for i in range(ini_num_population):
            p.append(PostDisasterLongTermSolution(unrepaired_bridge, node_df,
                                                  arc_df, bridge_df,
                                                  bridge_damage_value, network,
                                                  pm, all_ipw, path_adt))

        first_front = nsga2.run(p, population_size, num_generation)

        # output the NSGA result
        order = len(first_front) - 1
        obj0 = first_front[order].objectives[0]
        obj1 = first_front[order].objectives[1]
        repair_schedule = first_front[order].sch[obj0, obj1]

        # extract the time just finishing repairing a bridge
        ending_time = []
        bridge_set = []
        for key, value in repair_schedule.items():
            bridge_set.append(key)
            ending_time.append(value[1])

        # output the optimal solution of bridge repair schedule based on NSGA
        bridge_recovery = pd.DataFrame(
            {"Bridge ID": bridge_set, "Ending Time": ending_time})
        self.set_result_csv_data(
            "optimal_solution_of_bridge_repair_schedule", bridge_recovery,
            name="optimal_solution_of_bridge_repair_schedule", source="dataframe")

        network = TransportationRecoveryUtil.nw_reconstruct(node_df, arc_df, adt_data)

        temp_bridge_damage_value = copy.deepcopy(bridge_damage_value)

        # record the  non-recurrence ending time of bridges
        schedule_time = []
        end = {}
        for key, value in repair_schedule.items():
            if value[1] not in schedule_time:
                schedule_time.append(value[1])
            end[key] = value[1]
        schedule_time.sort()

        # label whether the bridge has been repaired
        bridge_repair = list(repair_schedule.keys())
        fg = {}
        for bridge in bridge_repair:
            fg[bridge] = 0

        # calculate the transportation network efficiency
        efficiency = []
        for ii in range(len(schedule_time)):

            # update the damage status of bridge
            for bridge in bridge_repair:
                if fg[bridge] == 0:
                    if end[bridge] <= schedule_time[ii]:
                        temp_bridge_damage_value[bridge] = 0
                        fg[bridge] = 1

            for i in range(len(arc_df)):
                nod1 = node_df.loc[node_df['ID'] == arc_df['fromnode'][i], 'guid'].values[0]
                nod2 = node_df.loc[node_df['ID'] == arc_df['tonode'][i], 'guid'].values[0]
                network.edges[nod1, nod2]['Damage_Status'] = 0

            for key, val in temp_bridge_damage_value.items():
                linknwid = bridge_df.loc[bridge_df['guid'] == key, 'linkID'].values[0]

                nod_id1 = arc_df[arc_df['id'] == linknwid]['fromnode'].values[0]
                nod1 = node_df.loc[node_df['ID'] == nod_id1, 'guid'].values[0]

                nod_id2 = arc_df[arc_df['id'] == linknwid]['tonode'].values[0]
                nod2 = node_df.loc[node_df['ID'] == nod_id2, 'guid'].values[0]

                network.edges[nod1, nod2]['Damage_Status'] = val

            # calculate different travel efficiency based on different
            # performance metrics
            current_te = None
            if pm == 1:
                current_te = TransportationRecoveryUtil.traveltime_freeflow(network)

            elif pm == 0:
                current_te = WIPW.tipw_index(network, all_ipw, path_adt)

            efficiency.append(current_te)

        # normalize the efficiency at each time of repairing a bridge
        try:
            maxe = max(efficiency)
            for i in range(len(efficiency)):
                efficiency[i] = efficiency[i] / maxe
        except ValueError as e:
            print(e)

        # output the recovery trajectory
        recovery_trajectory = pd.DataFrame(
            {"Ending Time": schedule_time, "Travel Efficiency": efficiency})
        self.set_result_csv_data("overall_transportation_recovery_trajectory", recovery_trajectory,
                                 name="overall_transportation_recovery_trajectory", source="dataframe")

        return None

    def get_spec(self):
        """Get specifications of the transportation recovery model.

        Returns:
            obj: A JSON object of specifications of the transportation recovery model.

        """
        return {
            'name': 'transportation-recovery',
            'description': 'transportation recovery model',
            'input_parameters': [
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
                {
                    'id': 'pm',
                    'required': True,
                    'description': 'transportation performance metrics 0: WIPW,  1:Free flow travel time',
                    'type': int
                },
                {
                    'id': 'ini_num_population',
                    'required': True,
                    'description': 'ini_num_population: 5 or 50',
                    'type': int
                },
                {
                    'id': 'population_size',
                    'required': True,
                    'description': 'population_size: 3 or 30',
                    'type': int
                },
                {
                    'id': 'num_generation',
                    'required': True,
                    'description': 'num_generation: 2 or 250',
                    'type': int
                },
                {
                    'id': 'mutation_rate',
                    'required': True,
                    'description': '0.1',
                    'type': float
                },
                {
                    'id': 'crossover_rate',
                    'required': True,
                    'description': '1.0',
                    'type': float
                }
            ],
            'input_datasets': [
                {
                    'id': 'nodes',
                    'required': True,
                    'description': 'road nodes',
                    'type': ['ergo:roadNetwork'],
                },
                {
                    'id': 'links',
                    'required': True,
                    'description': 'road links',
                    'type': ['ergo:roadNetwork'],
                },
                {
                    'id': 'bridges',
                    'required': True,
                    'description': 'bridges',
                    'type': ['ergo:bridges', 'ergo:bridgesVer2', 'ergo:bridgesVer3'],
                },
                {
                    'id': 'bridge_damage_value',
                    'required': True,
                    'description': '',
                    'type': ['incore:bridgeDamageValue']
                },
                {
                    'id': 'unrepaired_bridge',
                    'required': True,
                    'description': '',
                    'type': ['incore:unrepairedBridge']
                },
                {
                    'id': 'ADT',
                    'required': True,
                    'description': '',
                    'type': ['incore:ADT']
                }

            ],
            'output_datasets': [
                {
                    'id': 'optimal_solution_of_bridge_repair_schedule',
                    'description': 'List the Bridge id and its ending repair time.',
                    'type': 'incore:transportationRepairSchedule'
                },
                {
                    'id': 'overall_transportation_recovery_trajectory',
                    'description': 'shows the overall recovery trajectory of the ' +
                                   'transportation system. List the ending time and ' +
                                   'travel efficiency for the whole network.',
                    'type': 'incore:transportationRecovery'
                }
            ]
        }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import concurrent.futures
from itertools import repeat

from pyincore import AnalysisUtil, GeoUtil
from pyincore import BaseAnalysis, HazardService, FragilityService
from pyincore.analyses.nonstructbuildingdamage.nonstructbuildingutil import \
    NonStructBuildingUtil
from pyincore.models.dfr3curve import DFR3Curve


class NonStructBuildingDamage(BaseAnalysis):
    """Computes non-structural structural building damage for an earthquake hazard.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        self.hazardsvc = HazardService(incore_client)
        self.fragilitysvc = FragilityService(incore_client)

        super(NonStructBuildingDamage, self).__init__(incore_client)

    def run(self):
        """Executes building damage analysis."""
        # Building dataset
        building_set = self.get_input_dataset("buildings").get_inventory_reader()

        # get input hazard
        hazard, hazard_type, hazard_dataset_id = self.create_hazard_object_from_input_params()

        # set Default Fragility key
        fragility_key_as = self.get_parameter("fragility_key_as")
        if fragility_key_as is None:
            self.set_parameter("fragility_key_as", NonStructBuildingUtil.DEFAULT_FRAGILITY_KEY_AS)

        fragility_key_ds = self.get_parameter("fragility_key_ds")
        if fragility_key_ds is None:
            self.set_parameter("fragility_key_ds", NonStructBuildingUtil.DEFAULT_FRAGILITY_KEY_DS)

        # Set Default Hazard Uncertainty
        use_hazard_uncertainty = self.get_parameter("use_hazard_uncertainty")
        if use_hazard_uncertainty is None:
            self.set_parameter("use_hazard_uncertainty", False)

        # Set Default Liquefaction
        use_liquefaction = self.get_parameter("use_liquefaction")
        if use_liquefaction is None:
            self.set_parameter("use_liquefaction", False)

        user_defined_cpu = 1

        if not self.get_parameter("num_cpu") is None and self.get_parameter("num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self, len(building_set), user_defined_cpu)

        avg_bulk_input_size = int(len(building_set) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(building_set)

        while count < len(inventory_list):
            inventory_args.append(inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        (ds_results, damage_results) = self.building_damage_concurrent_future(self.building_damage_analysis_bulk_input,
                                                                              num_workers,
                                                                              inventory_args,
                                                                              repeat(hazard),
                                                                              repeat(hazard_type),
                                                                              repeat(hazard_dataset_id))

        self.set_result_csv_data("result", ds_results, name=self.get_parameter("result_name"))
        self.set_result_json_data("damage_result",
                                  damage_results,
                                  name=self.get_parameter("result_name") + "_additional_info")
        return True

    def building_damage_concurrent_future(self, function_name, num_workers, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            num_workers (int): Maximum number workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            dict: An ordered dictionary with building damage values.
            dict: An ordered dictionary with building data/metadata.

        """
        output_ds = []
        output_dmg = []
        with concurrent.futures.ProcessPoolExecutor(max_workers=num_workers) as executor:
            for ret1, ret2 in executor.map(function_name, *args):
                output_ds.extend(ret1)
                output_dmg.extend(ret2)

        return output_ds, output_dmg

    def building_damage_analysis_bulk_input(self, buildings, hazard, hazard_type, hazard_dataset_id):
        """Run analysis for multiple buildings.

        Args:
            buildings (list): Multiple buildings from input inventory set.
            hazard (obj): Hazard object.
            hazard_type (str): Hazard type.
            hazard_dataset_id (str): Hazard dataset id.

        Returns:
            dict: An ordered dictionary with building damage values.
            dict: An ordered dictionary with building data/metadata.

        """
        # read static parameters from object self
        liq_geology_dataset_id = self.get_parameter("liq_geology_dataset_id")
        use_liquefaction = self.get_parameter("use_liquefaction")
        use_hazard_uncertainty = self.get_parameter("use_hazard_uncertainty")

        # get allowed demand types for the hazard type
        allowed_demand_types = [item["demand_type"].lower() for item in self.hazardsvc.get_allowed_demands(
            hazard_type)]

        building_results = []
        damage_results = []
        fragility_sets_as = self.fragilitysvc.match_inventory(self.get_input_dataset("dfr3_mapping_set"), buildings,
                                                              self.get_parameter("fragility_key_as"))
        fragility_sets_ds = self.fragilitysvc.match_inventory(self.get_input_dataset("dfr3_mapping_set"), buildings,
                                                              self.get_parameter("fragility_key_ds"))
        values_payload_as = []
        values_payload_ds = []
        values_payload_liq = []
        mapped_buildings = []
        unmapped_buildings = []
        liquefaction_resp = None
        for building in buildings:
            if building["id"] in fragility_sets_as and building["id"] in fragility_sets_ds:
                fragility_set_as = fragility_sets_as[building["id"]]
                fragility_set_ds = fragility_sets_ds[building["id"]]
                location = GeoUtil.get_location(building)
                loc = str(location.y) + "," + str(location.x)

                # Acceleration-Sensitive
                demands_as, units_as, _ = AnalysisUtil.get_hazard_demand_types_units(building, fragility_set_as,
                                                                                     hazard_type, allowed_demand_types)
                value_as = {
                    "demands": demands_as,
                    "units": units_as,
                    "loc": loc
                }
                values_payload_as.append(value_as)

                # Drift-Sensitive
                demands_ds, units_ds, _ = AnalysisUtil.get_hazard_demand_types_units(building, fragility_set_ds,
                                                                                     hazard_type, allowed_demand_types)
                value_ds = {
                    "demands": demands_ds,
                    "units": units_ds,
                    "loc": loc
                }
                values_payload_ds.append(value_ds)

                # liquefaction
                if use_liquefaction:
                    value_liq = {
                        "demands": ["pgd"],  # implied...
                        "units": ["in"],
                        "loc": loc
                    }
                    values_payload_liq.append(value_liq)

                mapped_buildings.append(building)
            else:
                unmapped_buildings.append(building)

        del buildings

        # get hazard values and liquefaction
        if hazard_type == 'earthquake':
            hazard_resp_as = hazard.read_hazard_values(values_payload_as, self.hazardsvc)
            hazard_resp_ds = hazard.read_hazard_values(values_payload_ds, self.hazardsvc)

            # adjust dmg probability for liquefaction
            if use_liquefaction:
                if liq_geology_dataset_id is not None:
                    liquefaction_resp = self.hazardsvc.post_liquefaction_values(hazard_dataset_id,
                                                                                liq_geology_dataset_id,
                                                                                values_payload_liq)
                else:
                    raise ValueError('Hazard does not support liquefaction! Check to make sure you defined the '
                                     'liquefaction portion of your scenario earthquake.')
        else:
            raise ValueError("The provided hazard type is not supported yet by this analysis")

        # calculate LS and DS
        for i, building in enumerate(mapped_buildings):
            dmg_probability_as = {"LS_0": None, "LS_1": None, "LS_2": None}
            dmg_interval_as = {"DS_0": None, "DS_1": None, "DS_2": None, "DS_3": None}
            dmg_probability_ds = {"LS_0": None, "LS_1": None, "LS_2": None}
            dmg_interval_ds = {"DS_0": None, "DS_1": None, "DS_2": None, "DS_3": None}
            fragility_set_as = fragility_sets_as[building["id"]]
            fragility_set_ds = fragility_sets_ds[building["id"]]

            # TODO this value needs to come from the hazard service
            # adjust dmg probability for hazard uncertainty
            if use_hazard_uncertainty:
                raise ValueError('Uncertainty has not yet been implemented!')

            ###############
            # AS
            if isinstance(fragility_set_as.fragility_curves[0], DFR3Curve):
                hazard_vals_as = AnalysisUtil.update_precision_of_lists(hazard_resp_as[i]["hazardValues"])
                demand_types_as = hazard_resp_as[i]["demands"]
                demand_units_as = hazard_resp_as[i]["units"]
                hval_dict_as = dict()
                for j, d in enumerate(fragility_set_as.demand_types):
                    hval_dict_as[d] = hazard_vals_as[j]
                if not AnalysisUtil.do_hazard_values_have_errors(hazard_resp_as[i]["hazardValues"]):
                    building_args = fragility_set_as.construct_expression_args_from_inventory(building)
                    dmg_probability_as = fragility_set_as. \
                        calculate_limit_state(hval_dict_as, inventory_type="building",
                                              **building_args)
                    # adjust dmg probability for liquefaction
                    if use_liquefaction:
                        if liq_geology_dataset_id is not None:
                            liquefaction_dmg = AnalysisUtil.update_precision_of_lists(liquefaction_resp[i][
                                                                                          "groundFailureProb"])
                            dmg_probability_as = AnalysisUtil.update_precision_of_dicts(
                                NonStructBuildingUtil.adjust_damage_for_liquefaction(dmg_probability_as,
                                                                                     liquefaction_dmg))
                    dmg_interval_as = fragility_set_ds.calculate_damage_interval(dmg_probability_as,
                                                                                 hazard_type=hazard_type,
                                                                                 inventory_type="building")
            else:
                raise ValueError("One of the fragilities is in deprecated format. This should not happen. If you are "
                                 "seeing this please report the issue.")

            ###############
            # DS
            if isinstance(fragility_set_ds.fragility_curves[0], DFR3Curve):
                hazard_vals_ds = AnalysisUtil.update_precision_of_lists(hazard_resp_ds[i]["hazardValues"])
                demand_types_ds = hazard_resp_ds[i]["demands"]
                demand_units_ds = hazard_resp_ds[i]["units"]
                hval_dict_ds = dict()
                for j, d in enumerate(fragility_set_ds.demand_types):
                    hval_dict_ds[d] = hazard_vals_ds[j]

                if not AnalysisUtil.do_hazard_values_have_errors(hazard_resp_ds[i]["hazardValues"]):
                    building_args = fragility_set_ds.construct_expression_args_from_inventory(building)
                    dmg_probability_ds = fragility_set_ds. \
                        calculate_limit_state(hval_dict_ds, inventory_type="building",
                                              **building_args)
                    # adjust dmg probability for liquefaction
                    if use_liquefaction:
                        if liq_geology_dataset_id is not None:
                            liquefaction_dmg = AnalysisUtil.update_precision_of_lists(
                                liquefaction_resp[i]["groundFailureProb"])
                            dmg_probability_ds = AnalysisUtil.update_precision_of_dicts(
                                NonStructBuildingUtil.adjust_damage_for_liquefaction(dmg_probability_ds,
                                                                                     liquefaction_dmg))
                    dmg_interval_ds = fragility_set_ds.calculate_damage_interval(dmg_probability_ds,
                                                                                 hazard_type=hazard_type,
                                                                                 inventory_type="building")
            else:
                raise ValueError("One of the fragilities is in deprecated format. This should not happen. If you are "
                                 "seeing this please report the issue.")

            # put results in dictionary
            # AS denotes acceleration-sensitive fragility assigned to the building.
            # DS denotes drift-sensitive fragility assigned to the building.
            building_result = dict()
            building_result['guid'] = building['properties']['guid']
            building_result['AS_LS_0'] = dmg_probability_as['LS_0']
            building_result['AS_LS_1'] = dmg_probability_as['LS_1']
            building_result['AS_LS_2'] = dmg_probability_as['LS_2']
            building_result['AS_DS_0'] = dmg_interval_as['DS_0']
            building_result['AS_DS_1'] = dmg_interval_as['DS_1']
            building_result['AS_DS_2'] = dmg_interval_as['DS_2']
            building_result['AS_DS_3'] = dmg_interval_as['DS_3']
            building_result['DS_LS_0'] = dmg_probability_ds['LS_0']
            building_result['DS_LS_1'] = dmg_probability_ds['LS_1']
            building_result['DS_LS_2'] = dmg_probability_ds['LS_2']
            building_result['DS_DS_0'] = dmg_interval_ds['DS_0']
            building_result['DS_DS_1'] = dmg_interval_ds['DS_1']
            building_result['DS_DS_2'] = dmg_interval_ds['DS_2']
            building_result['DS_DS_3'] = dmg_interval_ds['DS_3']
            hazard_exposure_as = AnalysisUtil.get_exposure_from_hazard_values(hazard_vals_as, hazard_type)
            hazard_exposure_ds = AnalysisUtil.get_exposure_from_hazard_values(hazard_vals_ds, hazard_type)
            haz_expose = NonStructBuildingUtil.determine_haz_exposure(hazard_exposure_as, hazard_exposure_ds)
            building_result['haz_expose'] = haz_expose

            # put damage results in dictionary
            damage_result = dict()
            damage_result['guid'] = building['properties']['guid']
            damage_result['fragility_id_as'] = fragility_set_as.id
            damage_result['demandtypes_as'] = demand_types_as
            damage_result['demandunits_as'] = demand_units_as
            damage_result['fragility_id_ds'] = fragility_set_ds.id
            damage_result['demandtypes_ds'] = demand_types_ds
            damage_result['demandunits_ds'] = demand_units_ds
            damage_result['hazardtype'] = hazard_type
            damage_result['hazardvals_as'] = hazard_vals_as
            damage_result['hazardvals_ds'] = hazard_vals_ds

            building_results.append(building_result)
            damage_results.append(damage_result)

        for building in unmapped_buildings:
            building_result = dict()
            building_result['guid'] = building['properties']['guid']

            damage_result = dict()
            damage_result['guid'] = building['properties']['guid']
            damage_result['fragility_id_as'] = None
            damage_result['demandtypes_as'] = None
            damage_result['demandunits_as'] = None
            damage_result['fragility_id_ds'] = None
            damage_result['demandtypes_ds'] = None
            damage_result['demandunits_ds'] = None
            damage_result['hazardtype'] = None
            damage_result['hazardvals_as'] = None
            damage_result['hazardvals_ds'] = None

            building_results.append(building_result)
            damage_results.append(damage_result)

        return building_results, damage_results

    def get_spec(self):
        """Get specifications of the building damage analysis.

        Returns:
            obj: A JSON object of specifications of the building damage analysis.

        """
        return {
            'name': 'building-damage',
            'description': 'building damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'hazard_type',
                    'required': False,
                    'description': 'Hazard Type (e.g. earthquake)',
                    'type': str
                },
                {
                    'id': 'hazard_id',
                    'required': False,
                    'description': 'Hazard ID',
                    'type': str
                },
                {
                    'id': 'fragility_key_as',
                    'required': False,
                    'description': 'Acceleration-Sensitive Fragility key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'fragility_key_ds',
                    'required': False,
                    'description': 'Drift-Sensitive Fragility key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'use_liquefaction',
                    'required': False,
                    'description': 'Use liquefaction',
                    'type': bool
                },
                {
                    'id': 'liq_geology_dataset_id',
                    'required': False,
                    'description': 'liquefaction geology dataset id, \
                        if use liquefaction, you have to provide this id',
                    'type': str
                },
                {
                    'id': 'use_hazard_uncertainty',
                    'required': False,
                    'description': 'Use hazard uncertainty',
                    'type': bool
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
            ],
            'input_hazards': [
                {
                    'id': 'hazard',
                    'required': False,
                    'description': 'Hazard object',
                    'type': ["earthquake"]
                },
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'building Inventory',
                    'type': ['ergo:buildingInventoryVer4'],
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'buildings',
                    'description': 'CSV file of damage states for building non-structural damage',
                    'type': 'ergo:nsBuildingInventoryDamageVer3'
                },
                {
                    'id': 'damage_result',
                    'parent_type': 'buildings',
                    'description': 'Json file with information about applied hazard value and fragility',
                    'type': 'incore:nsBuildingInventoryDamageSupplement'
                }
            ]
        }



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


"""
Water Facility Damage
"""

import concurrent.futures
import random
from itertools import repeat

from pyincore import BaseAnalysis, HazardService, FragilityService, GeoUtil, \
    AnalysisUtil
from pyincore.models.dfr3curve import DFR3Curve


class WaterFacilityDamage(BaseAnalysis):
    """Computes water facility damage for an earthquake tsunami, tornado, or hurricane exposure.

    """

    DEFAULT_EQ_FRAGILITY_KEY = "pga"
    DEFAULT_TSU_FRAGILITY_KEY = "Non-Retrofit inundationDepth Fragility ID Code"
    DEFAULT_LIQ_FRAGILITY_KEY = "pgd"

    def __init__(self, incore_client):
        # Create Hazard and Fragility service
        self.hazardsvc = HazardService(incore_client)
        self.fragilitysvc = FragilityService(incore_client)

        super(WaterFacilityDamage, self).__init__(incore_client)

    def run(self):
        """Performs Water facility damage analysis by using the parameters from the spec
        and creates an output dataset in csv format

        Returns:
            bool: True if successful, False otherwise
        """
        # Facility dataset
        inventory_set = self.get_input_dataset(
            "water_facilities").get_inventory_reader()

        # get input hazard
        hazard, hazard_type, hazard_dataset_id = self.create_hazard_object_from_input_params()

        user_defined_cpu = 1

        if not self.get_parameter("num_cpu") is None and self.get_parameter(
                "num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self,
                                                                 len(inventory_set),
                                                                 user_defined_cpu)

        avg_bulk_input_size = int(len(inventory_set) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = list(inventory_set)
        while count < len(inventory_list):
            inventory_args.append(
                inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        (ds_results, damage_results) = self.waterfacility_damage_concurrent_futures(
            self.waterfacilityset_damage_analysis_bulk_input, num_workers,
            inventory_args, repeat(hazard), repeat(hazard_type), repeat(hazard_dataset_id))

        self.set_result_csv_data("result", ds_results, name=self.get_parameter("result_name"))
        self.set_result_json_data("metadata",
                                  damage_results,
                                  name=self.get_parameter("result_name") + "_additional_info")

        return True

    def waterfacility_damage_concurrent_futures(self, function_name,
                                                parallel_processes,
                                                *args):
        """Utilizes concurrent.future module.

            Args:
                function_name (function): The function to be parallelized.
                parallel_processes (int): Number of workers in parallelization.
                *args: All the arguments in order to pass into parameter function_name.

            Returns:
                list: A list of ordered dictionaries with water facility damage values
                list: A list of ordered dictionaries with other water facility data/metadata


        """
        output_ds = []
        output_dmg = []

        with concurrent.futures.ProcessPoolExecutor(max_workers=parallel_processes) as executor:
            for ret1, ret2 in executor.map(function_name, *args):
                output_ds.extend(ret1)
                output_dmg.extend(ret2)

        return output_ds, output_dmg

    def waterfacilityset_damage_analysis_bulk_input(self, facilities, hazard, hazard_type, hazard_dataset_id):
        """Gets applicable fragilities and calculates damage

        Args:
            facilities (list): Multiple water facilities from input inventory set.
            hazard (object): A hazard object.
            hazard_type (str): A hazard type of the hazard exposure (earthquake, tsunami, tornado, or hurricane).
            hazard_dataset_id (str): An id of the hazard exposure.

        Returns:
            list: A list of ordered dictionaries with water facility damage values
            list: A list of ordered dictionaries with other water facility data/metadata
        """

        # Liquefaction related variables
        use_liquefaction = False
        liquefaction_available = False
        fragility_sets_liq = None
        liquefaction_resp = None
        geology_dataset_id = None
        liq_hazard_vals = None
        liq_demand_types = None
        liq_demand_units = None
        liquefaction_prob = None
        loc = None

        # Obtain the fragility key
        fragility_key = self.get_parameter("fragility_key")

        if fragility_key is None:
            if hazard_type == 'tsunami':
                fragility_key = self.DEFAULT_TSU_FRAGILITY_KEY
            elif hazard_type == 'earthquake':
                fragility_key = self.DEFAULT_EQ_FRAGILITY_KEY
            else:
                raise ValueError(
                    "Hazard type other than Earthquake and Tsunami are not currently supported.")

            self.set_parameter("fragility_key", fragility_key)

        # Obtain the fragility set
        fragility_sets = self.fragilitysvc.match_inventory(
            self.get_input_dataset("dfr3_mapping_set"), facilities, fragility_key)

        # Obtain the liquefaction fragility Key
        liquefaction_fragility_key = self.get_parameter(
            "liquefaction_fragility_key")

        if hazard_type == "earthquake":
            if self.get_parameter("use_liquefaction") is True:
                if liquefaction_fragility_key is None:
                    liquefaction_fragility_key = self.DEFAULT_LIQ_FRAGILITY_KEY

                use_liquefaction = self.get_parameter("use_liquefaction")

                # Obtain the geology dataset
                geology_dataset_id = self.get_parameter("liquefaction_geology_dataset_id")

                if geology_dataset_id is not None:
                    fragility_sets_liq = self.fragilitysvc.match_inventory(
                        self.get_input_dataset("dfr3_mapping_set"), facilities,
                        liquefaction_fragility_key)

                    if fragility_sets_liq is not None:
                        liquefaction_available = True

        # Determine whether to use hazard uncertainty
        uncertainty = self.get_parameter("use_hazard_uncertainty")

        # Setup fragility translation structures
        values_payload = []
        values_payload_liq = []
        unmapped_waterfacilities = []
        mapped_waterfacilities = []

        for facility in facilities:
            if facility["id"] in fragility_sets.keys():
                # Fill in generic details
                fragility_set = fragility_sets[facility["id"]]
                location = GeoUtil.get_location(facility)
                loc = str(location.y) + "," + str(location.x)
                demands = fragility_set.demand_types
                units = fragility_set.demand_units
                value = {
                    "demands": demands,
                    "units": units,
                    "loc": loc
                }
                values_payload.append(value)
                mapped_waterfacilities.append(facility)

                # Fill in liquefaction parameters
                if liquefaction_available and facility["id"] in fragility_sets_liq:
                    fragility_set_liq = fragility_sets_liq[facility["id"]]
                    demands_liq = fragility_set_liq.demand_types
                    units_liq = fragility_set_liq.demand_units
                    value_liq = {
                        "demands": demands_liq,
                        "units": units_liq,
                        "loc": loc
                    }
                    values_payload_liq.append(value_liq)
            else:
                unmapped_waterfacilities.append(facility)

        del facilities

        if hazard_type == 'earthquake' or 'tsunami':
            hazard_resp = hazard.read_hazard_values(values_payload, self.hazardsvc)
        else:
            raise ValueError("The provided hazard type is not supported yet by this analysis")

        # Check if liquefaction is applicable
        if liquefaction_available:
            liquefaction_resp = self.hazardsvc.post_liquefaction_values(hazard_dataset_id,
                                                                        geology_dataset_id,
                                                                        values_payload_liq)

        # Calculate LS and DS
        facility_results = []
        damage_results = []

        for i, facility in enumerate(mapped_waterfacilities):
            fragility_set = fragility_sets[facility["id"]]
            limit_states = dict()
            dmg_intervals = dict()

            # Setup conditions for the analysis
            hazard_std_dev = 0

            if uncertainty:
                hazard_std_dev = random.random()

            if isinstance(fragility_set.fragility_curves[0], DFR3Curve):
                hazard_vals = AnalysisUtil.update_precision_of_lists(hazard_resp[i]["hazardValues"])
                demand_types = hazard_resp[i]["demands"]
                demand_units = hazard_resp[i]["units"]

                hval_dict = dict()

                for j, d in enumerate(fragility_set.demand_types):
                    hval_dict[d] = hazard_vals[j]

                if not AnalysisUtil.do_hazard_values_have_errors(hazard_resp[i]["hazardValues"]):
                    facility_args = fragility_set.construct_expression_args_from_inventory(facility)
                    limit_states = \
                        fragility_set.calculate_limit_state(hval_dict,
                                                            std_dev=hazard_std_dev,
                                                            inventory_type='water_facility',
                                                            **facility_args)
                    # Evaluate liquefaction: if it is not none, then liquefaction is available
                    if liquefaction_resp is not None:
                        fragility_set_liq = fragility_sets_liq[facility["id"]]

                        if isinstance(fragility_set_liq.fragility_curves[0], DFR3Curve):
                            liq_hazard_vals = AnalysisUtil.update_precision_of_lists(liquefaction_resp[i]["pgdValues"])
                            liq_demand_types = liquefaction_resp[i]["demands"]
                            liq_demand_units = liquefaction_resp[i]["units"]
                            liquefaction_prob = liquefaction_resp[i]['liqProbability']

                            hval_dict_liq = dict()

                            for j, d in enumerate(fragility_set_liq.demand_types):
                                hval_dict_liq[d] = liq_hazard_vals[j]

                            facility_liq_args = fragility_set_liq.construct_expression_args_from_inventory(facility)
                            pgd_limit_states = \
                                fragility_set_liq.calculate_limit_state(
                                    hval_dict_liq, std_dev=hazard_std_dev, inventory_type="water_facility",
                                    **facility_liq_args)
                        else:
                            raise ValueError("One of the fragilities is in deprecated format. "
                                             "This should not happen If you are seeing this please report the issue.")

                        limit_states = AnalysisUtil.adjust_limit_states_for_pgd(limit_states, pgd_limit_states)

                    dmg_intervals = fragility_set.calculate_damage_interval(limit_states,
                                                                            hazard_type=hazard_type,
                                                                            inventory_type='water_facility')
            else:
                raise ValueError("One of the fragilities is in deprecated format. This should not happen. If you are "
                                 "seeing this please report the issue.")

            # TODO: ideally, this goes into a single variable declaration section

            facility_result = {'guid': facility['properties']['guid'], **limit_states, **dmg_intervals}
            facility_result['haz_expose'] = AnalysisUtil.get_exposure_from_hazard_values(hazard_vals, hazard_type)
            damage_result = dict()
            damage_result['guid'] = facility['properties']['guid']
            damage_result['fragility_id'] = fragility_set.id
            damage_result['demandtypes'] = demand_types
            damage_result['demandunits'] = demand_units
            damage_result['hazardtype'] = hazard_type
            damage_result['hazardvals'] = hazard_vals

            if use_liquefaction and fragility_sets_liq and geology_dataset_id:
                damage_result['liq_fragility_id'] = fragility_sets_liq[facility["id"]].id
                damage_result['liqdemandtypes'] = liq_demand_types
                damage_result['liqdemandunits'] = liq_demand_units
                damage_result['liqhazval'] = liq_hazard_vals
                damage_result['liqprobability'] = liquefaction_prob
            else:
                damage_result['liq_fragility_id'] = None
                damage_result['liqdemandtypes'] = None
                damage_result['liqdemandunits'] = None
                damage_result['liqhazval'] = None
                damage_result['liqprobability'] = None

            facility_results.append(facility_result)
            damage_results.append(damage_result)

        for facility in unmapped_waterfacilities:
            facility_result = dict()
            damage_result = dict()
            facility_result['guid'] = facility['properties']['guid']
            damage_result['guid'] = facility['properties']['guid']
            damage_result['fragility_id'] = None
            damage_result['demandtypes'] = None
            damage_result['demandunits'] = None
            damage_result['hazardtype'] = None
            damage_result['hazardvals'] = None
            damage_result['liq_fragility_id'] = None
            damage_result['liqdemandtypes'] = None
            damage_result['liqdemandunits'] = None
            damage_result['liqhazval'] = None
            damage_result['liqprobability'] = None

            facility_results.append(facility_result)
            damage_results.append(damage_result)

        return facility_results, damage_results

    def get_spec(self):
        return {
            'name': 'water-facility-damage',
            'description': 'water facility damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': False,
                    'description': 'result dataset name',
                    'type': str
                },
                {
                    'id': 'hazard_type',
                    'required': False,
                    'description': 'Hazard Type (e.g. earthquake)',
                    'type': str
                },
                {
                    'id': 'hazard_id',
                    'required': False,
                    'description': 'Hazard ID',
                    'type': str
                },
                {
                    'id': 'fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'use_liquefaction',
                    'required': False,
                    'description': 'Use liquefaction',
                    'type': bool
                },

                {
                    'id': 'liquefaction_geology_dataset_id',
                    'required': False,
                    'description': 'Liquefaction geology/susceptibility dataset id. '
                                   'If not provided, liquefaction will be ignored',
                    'type': str
                },
                {
                    'id': 'liquefaction_fragility_key',
                    'required': False,
                    'description': 'Fragility key to use in liquefaction mapping dataset',
                    'type': str
                },
                {
                    'id': 'use_hazard_uncertainty',
                    'required': False,
                    'description': 'Use hazard uncertainty',
                    'type': bool
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
            ],
            'input_hazards': [
                {
                    'id': 'hazard',
                    'required': False,
                    'description': 'Hazard object',
                    'type': ["earthquake", "tsunami"]
                },
            ],
            'input_datasets': [
                {
                    'id': 'water_facilities',
                    'required': True,
                    'description': 'Water Facility Inventory',
                    'type': ['ergo:waterFacilityTopo'],
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                }
            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'water_facilities',
                    'description': 'A csv file with limit state probabilities and damage states '
                                   'for each water facility',
                    'type': 'ergo:waterFacilityDamageVer6'
                },
                {
                    'id': 'metadata',
                    'parent_type': 'water_facilities',
                    'description': 'additional metadata in json file about applied hazard value and '
                                   'fragility',
                    'type': 'incore:waterFacilityDamageSupplement'
                }
            ]
        }



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import pandas as pd
from pyincore import BaseAnalysis


class SocialVulnerability(BaseAnalysis):
    """This analysis computes a social vulnerability score for per associated zone in census data.

    The computation extracts zoning and a social vulnerability score obtained by computing demographic features of
    interest against national average values.

    The output of the computation is a dataset CSV format.

    Contributors
        | Science: Elaina Sutley, Amin Enderami
        | Implementation:  Amin Enderami, Santiago Núñez-Corrales, and NCSA IN-CORE Dev Team

    Related publications

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(SocialVulnerability, self).__init__(incore_client)

    def run(self):
        """Execute the social vulnerability analysis using known parameters."""
        df_navs = pd.DataFrame(self.get_input_dataset('national_vulnerability_feature_averages').get_csv_reader())

        df_dem = pd.DataFrame(self.get_input_dataset('social_vulnerability_demographic_factors').get_csv_reader())

        # Make sure data types match
        df_dem["factor_white_nonHispanic"] = df_dem["factor_white_nonHispanic"].astype(float)
        df_dem["factor_owner_occupied"] = df_dem["factor_owner_occupied"].astype(float)
        df_dem["factor_earning_higher_than_national_poverty_rate"] =\
            df_dem["factor_earning_higher_than_national_poverty_rate"].astype(float)
        df_dem["factor_over_25_with_high_school_diploma_or_higher"] =\
            df_dem["factor_over_25_with_high_school_diploma_or_higher"].astype(float)
        df_dem["factor_without_disability_age_18_to_65"] =\
            df_dem["factor_without_disability_age_18_to_65"].astype(float)

        self.social_vulnerability_model(df_navs, df_dem)

    def social_vulnerability_model(self, df_navs, df_dem):
        """

        Args:
            df_navs (pd.DataFrame): dataframe containing the national average values for vulnerability factors
            df_dem (pd.DataFrame): dataframe containing demographic factors required for the vulnerability score

        Returns:

        """

        # Compute the social vulnerability index
        df_sv = self.compute_svs(df_dem, df_navs)

        # Save into a CSV file
        result_name = self.get_parameter("result_name")
        self.set_result_csv_data("sv_result", df_sv,
                                 name=result_name,
                                 source="dataframe")

    @staticmethod
    def compute_svs(df, df_navs):
        """ Computation of the social vulnerability score and corresponding zoning

        Args:
            df (pd.DataFrame): dataframe for the census geographic unit of interest
            df_navs (pd.DataFrame): dataframe containing national average values

        Returns:
            pd.DataFrame: Social vulnerability score and corresponding zoning data
        """
        navs = df_navs['average'].astype(float).array

        df['R1'] = df['factor_white_nonHispanic'] / navs[0]
        df['R2'] = df['factor_owner_occupied'] / navs[1]
        df['R3'] = df['factor_earning_higher_than_national_poverty_rate'] / navs[2]
        df['R4'] = df['factor_over_25_with_high_school_diploma_or_higher'] / navs[3]
        df['R5'] = df['factor_without_disability_age_18_to_65'] / navs[4]
        df['SVS'] = df.apply(lambda row: (row['R1'] + row['R2'] + row['R3'] + row['R4'] + row['R5']) / 5, axis=1)

        maximum_nav = 1/navs
        std = abs(1 - (sum(maximum_nav) / len(maximum_nav))) / 3

        lb_2 = 1 - 1.5*std
        lb_1 = 1 - 0.5*std
        ub_1 = 1 + 0.5*std
        ub_2 = 1 + 1.5*std

        zones = []

        for svs in df['SVS'].tolist():
            if svs < lb_2:
                new_zone = 'High Vulnerable (zone5)'
            elif svs < lb_1:
                new_zone = 'Medium to High Vulnerable (zone4)'
            elif svs < ub_1:
                new_zone = 'Medium Vulnerable (zone3)'
            elif svs < ub_2:
                new_zone = 'Medium to Low Vulnerable (zone2)'
            elif svs > ub_2:
                new_zone = 'Low Vulnerable (zone1)'
            else:
                new_zone = 'No Data'
            zones.append(new_zone)

        df['zone'] = zones
        df = df.sort_values(by="GEO_ID")

        return df

    def get_spec(self):
        """Get specifications of the housing serial recovery model.

        Returns:
            obj: A JSON object of specifications of the social vulnerability model.

        """
        return {
            'name': 'social-vulnerability',
            'description': 'Social vulnerability score model',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'Result CSV dataset name',
                    'type': str
                },
            ],
            'input_datasets': [
                {
                    'id': 'national_vulnerability_feature_averages',
                    'required': True,
                    'description': 'A csv file with national vulnerability feature averages',
                    'type': ['incore:socialVulnerabilityFeatureAverages']
                },
                {
                    'id': 'social_vulnerability_demographic_factors',
                    'required': True,
                    'description': 'A csv file with social vulnerability demographic factors for a given geographic '
                                   'type',
                    'type': ['incore:socialVulnerabilityDemFactors']
                }
            ],
            'output_datasets': [
                {
                    'id': 'sv_result',
                    'parent_type': 'social_vulnerability_score',
                    'description': 'A csv file with zones containing demographic factors'
                                   'qualified by a social vulnerability score',
                    'type': 'incore:socialVulnerabilityScore'
                }
            ]
        }



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import pandas as pd
from pyincore import BaseAnalysis
import numpy as np


class CombinedWindWaveSurgeBuildingLoss(BaseAnalysis):
    """ 
    This analysis computes the building structural and content loss from wind, flood and surge-wave damage

    Contributors
        | Science: Omar Nofal, John W. van de Lindt, Trung Do, Guirong Yan, Sara Hamideh, Daniel Cox, Joel Dietrich
        | Implementation: Jiate Li, Chris Navarro and NCSA IN-CORE Dev Team

    Related publications
        Nofal, Omar & Lindt, John & Do, Trung & Yan, Guirong & Hamideh, Sara & Cox, Daniel & Dietrich, Joel. (2021).
        Methodology for Regional Multi-Hazard Hurricane Damage and Risk Assessment. Journal of Structural Engineering.
        147. 04021185. 10.1061/(ASCE)ST.1943-541X.0003144.
        
    Args:
        incore_client (IncoreClient): Service authentication.
    """

    def __init__(self, incore_client):
        super(CombinedWindWaveSurgeBuildingLoss, self).__init__(incore_client)

    def run(self):
        """Executes combined wind, wave, surge building loss analysis."""

        # Read buildings
        buildings = self.get_input_dataset("buildings").get_dataframe_from_shapefile()

        # Read Building wind damage
        wind_damage = self.get_input_dataset("wind_damage").get_dataframe_from_csv()

        # Read Building surge-wave damage
        surge_wave_damage = self.get_input_dataset("surge_wave_damage").get_dataframe_from_csv()

        # Read Building flood damage
        flood_damage = self.get_input_dataset("flood_damage").get_dataframe_from_csv()

        # Read cumulative replacement cost ratio of building content
        content_cost = self.get_input_dataset("content_cost").get_dataframe_from_csv()

        # Read cumulative replacement cost ratio of structural damage
        structure_cost = self.get_input_dataset("structural_cost").get_dataframe_from_csv()

        combined_loss = self.get_combined_loss(wind_damage, surge_wave_damage, flood_damage, buildings, content_cost,
                                               structure_cost)

        # Create the result dataset
        self.set_result_csv_data("result", combined_loss, self.get_parameter("result_name") + "_loss",
                                 "dataframe")

        return True
    
    def get_combined_loss(self, wind_dmg: pd.DataFrame, sw_dmg: pd.DataFrame, flood_dmg: pd.DataFrame, buildings:
                          pd.DataFrame, content_cost: pd.DataFrame, structure_cost: pd.DataFrame):
        """Calculates structural and content loss from wind, surge-wave and flood damage

                Args:
                    wind_dmg (pd.DataFrame): Table of wind damage for the building inventory
                    sw_dmg (pd.DataFrame): Table of surge-wave damage for the building inventory
                    flood_dmg (pd.DataFrame): Table of flood damage for the building inventory
                    buildings (pd.DataFrame): Table of building attributes
                    content_cost (pd.DataFrame): Table of content cost ratios for each archetype  
                    structure_cost (pd.DataFrame): Table of structural cost ratio for each archetype and loss type 

                Returns:
                    pd.DataFrame: An table of structural and content loss for each building

                """
        # Rename the columns so there are no overlapping names among the hazard damage states
        flood_dmg.rename(columns={'LS_0': 'f_LS_0', 'LS_1': 'f_LS_1', 'LS_2': 'f_LS_2', 'DS_0': 'f_DS_0',
                                  'DS_1': 'f_DS_1', 'DS_2': 'f_DS_2', 'DS_3': 'f_DS_3', 'haz_expose': 'f_haz_expose'},
                         inplace=True)
        
        sw_dmg.rename(columns={'LS_0': 'sw_LS_0', 'LS_1': 'sw_LS_1', 'LS_2': 'sw_LS_2', 'DS_0': 'sw_DS_0',
                               'DS_1': 'sw_DS_1', 'DS_2': 'sw_DS_2', 'DS_3': 'sw_DS_3', 'haz_expose': 
                                   'sw_haz_expose'}, inplace=True)

        wind_dmg.rename(columns={'LS_0': 'w_LS_0', 'LS_1': 'w_LS_1', 'LS_2': 'w_LS_2', 'DS_0': 'w_DS_0',
                                 'DS_1': 'w_DS_1', 'DS_2': 'w_DS_2', 'DS_3': 'w_DS_3', 'haz_expose': 'w_haz_expose'},
                        inplace=True)

        # Combine all three sets of damages into a single data frame
        combined_df = pd.merge(pd.merge(wind_dmg, sw_dmg, on='guid'), flood_dmg, on='guid')

        # Add the flood archetype to the combined damage since this is used to find loss multipliers
        new_combined_df = pd.merge(combined_df, buildings[["guid", "arch_flood"]], on="guid")

        # Create a result data frame for the loss calculations
        loss_df = new_combined_df[['guid']].copy()

        # Compute content and structural loss
        loss_df['cont_loss'] = new_combined_df.apply(lambda row: np.dot(row[['f_DS_0', 'f_DS_1', 'f_DS_2']],
                                                                        content_cost.iloc[int(row['arch_flood']-1),
                                                                        1:4]), axis=1)

        loss_df['roof_loss'] = new_combined_df.apply(lambda row: np.dot(row[['w_DS_0', 'w_DS_1', 'w_DS_2', 'w_DS_3']],
                                                                        np.array([0.15, 0.5, 0.75, 1])) *
                                                     structure_cost.loc[int(row['arch_flood']-1), "Roofing"], axis=1)

        loss_df['ff_loss'] = new_combined_df.apply(lambda row: np.dot(row[['sw_DS_0', 'sw_DS_1', 'sw_DS_2',
                                                                           'sw_DS_3']], np.array([0.1, 0.5, 0.75, 1])) *
                                                   structure_cost.loc[int(row['arch_flood'] - 1),
                                                   "Flooring and Foundation"], axis=1)

        # Computes frame loss from the dominant hazard between wind and surge-wave intensities
        def compute_frame_loss(row):
            if row['w_DS_3'] >= row['sw_DS_3']:
                return np.dot(row[['w_DS_0', 'w_DS_1', 'w_DS_2', 'w_DS_3']],np.array([0.25, 0.5, 0.75, 1])) * \
                    structure_cost.loc[int(row['arch_flood'] - 1), "Wood Framing"]
            elif row['sw_DS_3'] > row['w_DS_3']:
                return np.dot(row[['sw_DS_0', 'sw_DS_1', 'sw_DS_2', 'sw_DS_3']], np.array([0.25, 0.5, 0.75, 1])) * \
                    structure_cost.loc[int(row['arch_flood'] - 1), "Wood Framing"]
            else:
                # If one or both are nan, they can't be compared
                if pd.isnull(row['sw_DS_3']) and pd.isnull(row['w_DS_3']):
                    return 0
                elif not pd.isnull(row['sw_DS_3']):
                    return np.dot(row[['sw_DS_0', 'sw_DS_1', 'sw_DS_2', 'sw_DS_3']], np.array([0.25, 0.5, 0.75, 1])) * \
                                  structure_cost.loc[int(row['arch_flood'] - 1), "Wood Framing"]
                else:
                    return np.dot(row[['w_DS_0', 'w_DS_1', 'w_DS_2', 'w_DS_3']], np.array([0.25, 0.5, 0.75,
                                                                                           1])) * structure_cost.loc[
                        int(row['arch_flood'] - 1), "Wood Framing"]

        loss_df['frame_loss'] = new_combined_df.apply(lambda row: compute_frame_loss(row), axis=1)

        # Fill NA values with 0 otherwise we can't compute the total loss since empty are NaN values and won't add
        loss_df.fillna(0, inplace=True)
        loss_df['total_loss'] = loss_df.apply(lambda row: row['cont_loss'] + row['frame_loss'] + row[
            'roof_loss'] + row['ff_loss'], axis=1)

        return loss_df
    
    def get_spec(self):
        """Get specifications of the combined wind, wave, and surge building loss analysis.

        Returns:
            obj: A JSON object of specifications of the combined wind, wave, and surge building damage analysis.

        """
        return {
            'name': 'combined-wind-wave-surge-building-loss',
            'description': 'Combined wind wave and surge building loss analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventoryVer4', 'ergo:buildingInventoryVer5',
                             'ergo:buildingInventoryVer6', 'ergo:buildingInventoryVer7'],
                },
                {
                    'id': 'wind_damage',
                    'required': True,
                    'description': 'Wind damage result that has damage intervals in it',
                    'type': ['ergo:buildingDamageVer6']
                },
                {
                    'id': 'surge_wave_damage',
                    'required': True,
                    'description': 'Surge-wave damage result that has damage intervals in it',
                    'type': ['ergo:buildingDamageVer6']
                },
                {
                    'id': 'flood_damage',
                    'required': True,
                    'description': 'Flood damage result that has damage intervals in it',
                    'type': ['ergo:buildingDamageVer6']
                },
                {
                    'id': 'structural_cost',
                    'required': True,
                    'description': 'Structural cost ratio for each archetype and loss type',
                    'type': ['incore:structuralCostRatio']
                },
                {
                    'id': 'content_cost',
                    'required': True,
                    'description': 'Content cost ratio for each archetype and damage state',
                    'type': ['incore:contentCostRatio']
                }

            ],
            'output_datasets': [
                {
                    'id': 'result',
                    'parent_type': 'buildings',
                    'description': 'CSV file of building structural and content loss',
                    'type': 'incore:buildingLoss'
                }
            ]
        }



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


from pyincore.analyses.indp.indp import INDP
from pyincore.analyses.indp.dislocationutils import DislocationUtil
from pyincore.analyses.indp.indpcomponents import INDPComponents
from pyincore.analyses.indp.indpresults import INDPResults
from pyincore.analyses.indp.indputil import INDPUtil
from pyincore.analyses.indp.infrastructurearc import InfrastructureArc
from pyincore.analyses.indp.infrastructureinterdeparc import InfrastructureInterdepArc
from pyincore.analyses.indp.infrastructurenetwork import InfrastructureNetwork
from pyincore.analyses.indp.infrastructurenode import InfrastructureNode
from pyincore.analyses.indp.infrastructurespace import InfrastructureSpace
from pyincore.analyses.indp.infrastructureutil import InfrastructureUtil
